"""
    Solvers.ABC.CustomInferenceUtil.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
from typing import Callable, List
import numpy as np
import pandas as pd

from pyabc.acceptor import Acceptor
from pyabc.distance import Distance
from pyabc.epsilon import Epsilon
from pyabc.model import Model
from pyabc.parameters import Parameter
from pyabc.population import Particle
from pyabc.random_choice import fast_random_choice
from pyabc.random_variables import RV, Distribution
from pyabc.transition import ModelPerturbationKernel, Transition
from pyabc.inference_util.inference_util import (
    create_prior_pdf,
    create_transition_pdf,
    # create_weight_function,   # customized below
    # generate_valid_proposal,
    # evaluate_proposal,        # customized below
    # only_simulate_data_for_proposal
)

logger = logging.getLogger("ABC")

def create_weight_function(
    prior_pdf: Callable,
    transition_pdf: Callable,
) -> Callable:
    """Create a function that calculates a sample's importance weight.
    The weight is the prior divided by the transition density and the
    acceptance step weight.

    Parameters
    ----------
    prior_pdf: The prior density.
    transition_pdf: The transition density.

    Returns
    -------
    weight_function: The importance sample weight function.
    """

    def weight_function(m_ss, theta_ss, acceptance_weight: float):
        """Calculate total weight, from sampling and acceptance weight.

        Parameters
        ----------
        m_ss: The model sample.
        theta_ss: The parameter sample.
        acceptance_weight: The acceptance weight sample. In most cases 1.

        Returns
        -------
        weight: The total weight.
        """
        # prior and transition density (can be equal)
        prior_pd = prior_pdf(m_ss, theta_ss)
        transition_pd = transition_pdf(m_ss, theta_ss)
        # calculate weight
        weight = acceptance_weight * prior_pd / transition_pd
        if False:
            if np.isnan(weight):
                print("m_ss=", m_ss)
                print("theta_ss=", theta_ss)
                print("acceptance_weight=", acceptance_weight)
                print("prior_pd=", prior_pd)
                print("transition_pd=", transition_pd)
                raise ValueError("weight is NaN")
        return weight

    return weight_function

def generate_valid_proposal(
    t: int,
    m: np.ndarray,
    p: np.ndarray,
    model_prior: RV,
    parameter_priors: List[Distribution],
    model_perturbation_kernel: ModelPerturbationKernel,
    transitions: List[Transition],
):
    """Sample a parameter for a model.

    Parameters
    ----------
    t: Population index to generate for.
    m: Indices of alive models.
    p: Probabilities of alive models.
    model_prior: The model prior.
    parameter_priors: The parameter priors.
    model_perturbation_kernel: The model perturbation kernel.
    transitions: The transitions, one per model.

    Returns
    -------
    (m_ss, theta_ss): Model, parameter.
    """
    # first generation
    if t == 0:
        # sample from prior
        m_ss = int(model_prior.rvs())
        theta_ss = parameter_priors[m_ss].rvs()
        return m_ss, theta_ss

    # later generation
    # counter
    n_sample, n_sample_soft_limit = 0, 1000
    # sample until the prior density is positive
    while True:
        if len(m) > 1:
            index = fast_random_choice(p)
            m_s = m[index]
            m_ss = model_perturbation_kernel.rvs(m_s)
            # theta_s is None if the population m_ss has died out.
            # This can happen since the model_perturbation_kernel
            # can return a model nr which has died out.
            if m_ss not in m:
                continue
        else:
            # only one model
            m_ss = m[0]
        theta_ss = transitions[m_ss].rvs()

        # check if positive under prior
        # if model_prior.pmf(m_ss) * parameter_priors[m_ss].pdf(theta_ss) > 0:
        mp = model_prior.pmf(m_ss)
        pp = parameter_priors[m_ss].pdf(theta_ss)
        # print("mp, pp=", mp, pp)
        if mp * pp > 0:
            return m_ss, theta_ss

        # unhealthy sampling detection
        n_sample += 1
        if n_sample == n_sample_soft_limit:
            logger.warning(
                "Unusually many (model, parameter) samples have prior "
                "density zero. The transition might be inappropriate."
            )

def evaluate_proposal(
    m_ss: int,
    theta_ss: Parameter,
    t: int,
    models: List[Model],
    summary_statistics: Callable,
    distance_function: Distance,
    eps: Epsilon,
    acceptor: Acceptor,
    x_0: dict,
    weight_function: Callable,
    proposal_id: int,
) -> Particle:
    """Evaluate a proposed parameter.

    Parameters
    ----------
    m_ss, theta_ss: The proposed (model, parameter) sample.
    t: The current time.
    models: List of all models.
    summary_statistics:
        Function to compute summary statistics from model output.
    distance_function: The distance function.
    eps: The epsilon threshold.
    acceptor: The acceptor.
    x_0: The observed summary statistics.
    weight_function: Function by which to reweight the sample.
    proposal_id: Id of the transition kernel.

    Returns
    -------
    particle: A particle containing all information.

    Data for the given parameters theta_ss are simulated, summary statistics
    computed and evaluated.
    """
    # simulate, compute distance, check acceptance
    model_result = models[m_ss].accept(
        t, theta_ss, summary_statistics, distance_function, eps, acceptor, x_0
    )

    # compute acceptance weight
    if model_result.accepted:
        weight = weight_function(m_ss, theta_ss, model_result.weight)
        if np.isnan(weight):
            weight = 0
            model_result.accepted = False
    else:
        weight = 0

    return Particle(
        m=m_ss,
        parameter=theta_ss,
        weight=weight,
        sum_stat=model_result.sum_stat,
        distance=model_result.distance,
        accepted=model_result.accepted,
        preliminary=False,
        proposal_id=proposal_id,
    )

def only_simulate_data_for_proposal(
    m_ss: int,
    theta_ss: Parameter,
    t: int,
    models: List[Model],
    summary_statistics: Callable,
    weight_function: Callable,
    proposal_id: int,
) -> Particle:
    """Simulate data for parameters.

    Similar to `evaluate_proposal`, however here for the passed parameters
    only data are simulated, but no distances calculated or acceptance
    checked. That needs to be done post-hoc then, not checked here."""

    # simulate
    model_result = models[m_ss].summary_statistics(
        t, theta_ss, summary_statistics
    )

    # dummies for distance and weight, need to be recomputed later
    distance = np.inf
    acceptance_weight = 1.0

    # needs to be accepted in order to be forwarded by the sampler, and so
    #  as a single particle
    accepted = True

    # compute weight
    weight = weight_function(m_ss, theta_ss, acceptance_weight)
    if np.isnan(weight):
        weight = 0
        accepted = False

    return Particle(
        m=m_ss,
        parameter=theta_ss,
        weight=weight,
        sum_stat=model_result.sum_stat,
        distance=distance,
        accepted=accepted,
        preliminary=True,
        proposal_id=proposal_id,
    )

def create_simulate_function(
    t: int,
    model_probabilities: pd.DataFrame,
    model_perturbation_kernel: ModelPerturbationKernel,
    transitions: List[Transition],
    model_prior: RV,
    parameter_priors: List[Distribution],
    models: List[Model],
    summary_statistics: Callable,
    x_0: dict,
    distance_function: Distance,
    eps: Epsilon,
    acceptor: Acceptor,
    evaluate: bool = True,
    proposal_id: int = 0,
) -> Callable:
    """
    Create a simulation function which performs the sampling of parameters,
    simulation of data and acceptance checking, and which is then passed
    to the sampler.

    Parameters
    ----------
    t: The time index to simulate for.
    model_probabilities: The last generation's model probabilities.
    model_perturbation_kernel: The model perturbation kernel.
    transitions: The parameter transition kernels.
    model_prior: The model prior.
    parameter_priors: The parameter priors.
    models: List of all models.
    summary_statistics:
        Function to compute summary statistics from model output.
    x_0: The observed summary statistics.
    distance_function: The distance function.
    eps: The epsilon threshold.
    acceptor: The acceptor.
    evaluate:
        Whether to actually evaluate the sample. Should be True except for
        certain preliminary settings.
    proposal_id:
        Identifier for the proposal distribution.

    Returns
    -------
    simulate_one: callable
        Function that samples parameters, simulates data, and checks
        acceptance.

    .. note::
        For some of the samplers, the sampling function needs to be
        serialized in order to be transported to where the sampling
        happens. Therefore, the returned function should be light, and
        in particular not contain references to the ABCSMC class.
    """
    # cache model_probabilities to not query the database so often
    m = np.array(model_probabilities.index)
    p = np.array(model_probabilities.p)

    # create prior and transition densities for weight function
    prior_pdf = create_prior_pdf(
        model_prior=model_prior, parameter_priors=parameter_priors
    )
    if t == 0:
        transition_pdf = prior_pdf
    else:
        transition_pdf = create_transition_pdf(
            transitions=transitions,
            model_probabilities=model_probabilities,
            model_perturbation_kernel=model_perturbation_kernel,
        )

    # create weight function
    weight_function = create_weight_function(
        prior_pdf=prior_pdf, transition_pdf=transition_pdf
    )

    # simulation function
    def simulate_one():
        proposed = False
        for n in range(10):
            parameter = generate_valid_proposal(
                t=t,
                m=m,
                p=p,
                model_prior=model_prior,
                parameter_priors=parameter_priors,
                model_perturbation_kernel=model_perturbation_kernel,
                transitions=transitions,
            )
            if evaluate:
                particle = evaluate_proposal(
                    *parameter,
                    t=t,
                    models=models,
                    summary_statistics=summary_statistics,
                    distance_function=distance_function,
                    eps=eps,
                    acceptor=acceptor,
                    x_0=x_0,
                    weight_function=weight_function,
                    proposal_id=proposal_id,
                )
                proposed = True
                break
            else:
                particle = only_simulate_data_for_proposal(
                    *parameter,
                    t=t,
                    models=models,
                    summary_statistics=summary_statistics,
                    weight_function=weight_function,
                    proposal_id=proposal_id,
                )
                if particle.accepted:
                    proposed = True
                    break
                else:
                    # discard this proposal
                    continue

        assert proposed

        return particle

    return simulate_one



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Generate RDF from PDB
# Copyright (C) 2011 Stas Bevc, stas.bevc@cmm.ki.si
# www.sicmm.org

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>


from os import listdir
from math import sqrt, ceil, floor, pow, pi


# generate radial distribution function from
# particle positions stored in PDB files
def makeRDF(pdbdir, side, numbins, cm=False, numAT=0):
    
    maxbin = numbins # number of bins
    sideh = side/2.0
    dr = float(sideh/maxbin) # bin width
    hist = [0]*(maxbin+1) # initialize list
    rdf = {}
    count = 0
    pdbs = listdir(pdbdir) # list of files
    nstep = len(pdbs)
    
    print "Directory "+pdbdir+" has "+str(len(pdbs))+" files."
    
    # loop over pdb files
    for pdb in pdbs:
        if not pdb.endswith(".pdb"):
            nstep -= 1
            continue # skip other files
        count += 1
        print "Reading file ... "+pdb+" ("+ str(count) +"/"+ str(len(pdbs)) +")"
        
        # read atom coordinates from PDB
        atoms = []
        cmm = [0,0,0] # center of mass of molecule
        atc = 1 # atom counter
        lines = open(pdbdir+"/"+pdb)
        for line in lines:
            if line[0:4] != "ATOM":
                continue # skip other lines
            coords = map(float, (line[31:54]).split())
            
            if cm == True: # calculate center of mass
                # we assume masses of all particles are 1.0
                cmm[0] += coords[0]
                cmm[1] += coords[1]
                cmm[2] += coords[2]
                if atc < numAT:
                    atc += 1
                else:
                    atc = 1
                    cmm[0] /= numAT
                    cmm[1] /= numAT
                    cmm[2] /= numAT
                    
                    # fold coordinates
                    for i in range(3):
                        tmp = floor(cmm[i] * (1.0/side))
                        cmm[i] -= tmp * side
                    
                    atoms.append((cmm[0],cmm[1],cmm[2]))
                    cmm = [0,0,0]
            else: # no cm calculation
                atoms.append((coords[0], coords[1], coords[2]))
        
        # loop over particle pairs
        npart = len(atoms)
        print " looping over particle pairs (" +str(npart)+ "^2) ... "
        for i in range(npart):
            
            xi = (atoms[i])[0]
            yi = (atoms[i])[1]
            zi = (atoms[i])[2]
            
            for j in range(i+1, npart):
                xx = xi - (atoms[j])[0]
                yy = yi - (atoms[j])[1]
                zz = zi - (atoms[j])[2]
                
                # minimum image
                if (xx < -sideh):   xx = xx + side
                if (xx > sideh):    xx = xx - side
                if (yy < -sideh):   yy = yy + side
                if (yy > sideh):    yy = yy - side
                if (zz < -sideh):   zz = zz + side
                if (zz > sideh):    zz = zz - side
                
                # distance between i and j
                rd  = xx * xx + yy * yy + zz * zz
                rij = sqrt(rd)
                
                bin = int(ceil(rij/dr)) # determine in which bin the distance falls
                if (bin <= maxbin):
                    hist[bin] += 1
    
    # normalize
    print "Normalizing ... "
    phi = npart/pow(side, 3.0) # number density (N*V)
    norm = 2.0 * pi * dr * phi * nstep * npart
    
    for i in range(1, maxbin+1):
        rrr = (i - 0.5) * dr
        val = hist[i]/ norm / ((rrr * rrr) + (dr * dr) / 12.0)
        rdf.update({rrr:val})
    
    return rdf

#-------------------------------------------------------------------#

# write RDF into file
boxsize = 36.845
numbins = 384 # number of bins
cm = True # calculate RDF from center of mass of molecule
numAT = 4 # number of atoms in molecule
pdbsdir = "./pdbs/" # directory with PDB files
outfile = "./rdf.out"

rdf = makeRDF(pdbsdir, boxsize, numbins, cm, numAT)
print "Writing output file ... " +outfile
outfile = open(outfile, "w")
for r in sorted(rdf.iterkeys()): # sort before writing into file
    outfile.write("%15.8g %15.8g\n"%(r, rdf[r]))
outfile.close()






# coding: utf-8
"""

    StdoutRedirector.py

    Copyright (c) 2020, Masatsuyo Takahashi, KEK-PF

"""
import sys
import tkinter as Tk

class StdoutRedirector:
    def __init__(self, log_text):
        self.saved_stdout = sys.stdout
        self.log_text = log_text
        sys.stdout = self

    def write(self, string):
        try:
            if string is not None:
                self.log_text.insert(Tk.INSERT, string)
            self.log_text.see(Tk.END)
        except:
            # it seems that this can be called after destruction.
            pass

    def flush(self):
        pass

    def __del__(self):
        sys.stdout = self.saved_stdout
        # print('__del__ ok')



# coding: utf-8
"""
    SecTools.CorMap.CorMapTestUtils.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import os
from molass_legacy.KekLib.OurTkinter import Tk

class Manipulator:
    def __init__(self, restart_str=None):
        self.counter = -1
        self.restart_str = restart_str
        self.restarting = restart_str is not None
        self.prepare_out_folder()

    def prepare_out_folder(self):
        self.out_folder = "temp/figs"
        if not os.path.exists(self.out_folder):
            from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry
            mkdirs_with_retry(self.out_folder)

    def do_for_a_file(self, in_folder, uv_folder, plot):
        assert in_folder == uv_folder
        self.counter += 1
        if self.restarting:
            if in_folder.find(self.restart_str) >= 0:
                self.restarting = False
            else:
                print("skipping", in_folder)
                return True, None
        print(in_folder)
        agent = self.agent
        agent.fe.focus_force()
        agent.fe.delete( 0, Tk.END)
        agent.fe.insert( 0, in_folder )
        agent.on_in_folder_entry()
        file = os.path.join(self.out_folder, "fig-%03d" % self.counter)
        agent.save_the_figure(file)
        return True, None

    def manipulate(self, client, agent):
        from DataUtils import get_pytools_folder, serial_folder_walk
        self.agent = agent
        self.client = client
        pytools = get_pytools_folder()
        root_folder = os.path.join(pytools, "Data")
        serial_folder_walk(root_folder, self.do_for_a_file)



"""
    Batch.OptDataSetsProxy.py

    Copyright (c) 2023-2025, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from molass_legacy.Optimizer.OptDataSets import OptDataSets, get_dsets_impl

class DummyRgCurve:
    def __init__(self):
        poresize =  get_setting('poresize')
        if poresize is None:
            set_setting('poresize', 100)    # poresize will be used to compute dummy rg_params in SecTheory.SecEstimator.guess_initial_secparams

    def get_rgs_from_trs(self, trs):
        return np.ones(len(trs)) * np.nan

class OptDataSetsProxy(OptDataSets):
    def __init__(self, sd, corrected_sd):
        dsets = get_dsets_impl(sd, corrected_sd, rg_info=False)
        self.dsets = (dsets[0], DummyRgCurve(), dsets[2])
        D = dsets[0][1]
        E = sd.intensity_array[:,:,2].T
        self.weight_info = self.compute_weight_info(1/(E + D/100))
        self.E = E



# coding: utf-8
"""
    SmoothSpline.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""
import numpy                as np
from scipy.interpolate      import UnivariateSpline

DEBUG = False

class SmoothSpline:
    def __init__( self, x, y, w=None, curvature=False, sf_log_base=36, sf=None, is_guinier=True, with_smoother=False ):

        if DEBUG: print( 'SmoothSpline: x.shape[0]=', x.shape[0], ', x[0], x[-1]=', x[0], x[-1] )

        self.is_guinier = is_guinier

        yvar = np.var( y )
        if DEBUG: print( 'yvar=',  yvar, ', y.shape[0]=', y.shape[0] )
        if yvar == 0:
            yvar = 1e2
            if DEBUG:
                print( 'SmoothSpline: yvar(==0) is replaced to ', yvar )

        if sf is None:
            logvar = np.log( yvar * y.shape[0] )
            sf = np.exp( sf_log_base + logvar )
            if DEBUG: print( 'logvar=%.3g' % logvar, ', sf=%.3e' % sf )

        if with_smoother:
            from GaussianProcessDep import Smoother
            smoother = Smoother( x, y  )
            y_ = smoother( x )
        else:
            y_ = y

        # TODO: validity of min( 3, len(y)-1 )
        spline  = UnivariateSpline( x, y_, w=w, k=min( 3, len(y)-1 ), s=sf )
        if DEBUG: print( 'UnivariateSpline ok, curvature=', curvature )

        self.spline = spline

        if not curvature: return
        if DEBUG: print( 'derivatives' )

        yˈ  = spline.derivative(1)( x )
        yˈˈ = spline.derivative(2)( x )
        cy = yˈˈ / np.power( 1 + yˈ**2, 3/2 )

        self.yˈ  = yˈ
        self.yˈˈ = yˈˈ
        self.cy  = cy
        self.cv_array   = np.percentile( cy, [ 5, 20, 50, 80, 95 ] )
        cvmax = np.max( np.abs( self.cv_array[1:-1] ) )

        if DEBUG: print( 'cvmax=', cvmax, ', cv_array=', self.cv_array  )
        if cvmax > 100:
            # cvmax == 6.99 for IpI/exampleA_8mg.dat
            raise Exception( 'Unexpected cvmax=%g' % cvmax )

    def __call__( self, x ):
        return self.spline( x )



# coding: utf-8
"""
    EghMixture.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
import scipy.stats as stats
from scipy.optimize import fsolve
from sklearn.cluster import KMeans
import molass_legacy.KekLib.DebugPlot as plt

USE_MODE_AS_TR_INIT = True

def egh(x, H=1, tR=0, sigma=1.0, tau=0.0):
    x_  = x - tR
    s2  = 2 * sigma**2
    z   = s2 + tau*x_
    z_neg   = z <= 0
    z_pos   = z > 0

    zero_part = np.zeros( len(x) )[z_neg]
    posi_part = H * np.exp( -x_[z_pos]**2/( s2 + tau*x_[z_pos] ) )

    if tau >= 0:
        parts = [ zero_part, posi_part ]
    else:
        parts = [ posi_part, zero_part ]

    return np.hstack( parts )

"""
Kevin Lan, James W. Jorgenson, Journal of Chromatography A, 915 (2001) 1?13
"""
A0i = np.array([4, -6.293724, 9.232834, -11.342910, 9.123978, -4.173753, 0.827797])
e0 = np.poly1d(A0i[::-1])
"""
e0(th) = a0 + a1*th + a2*th**2 + ... + am*th**m
"""
A1i = np.array([0.75, 0.033807, -0.301080, 1.200371, -1.813317, 1.279318, -0.326582])
e1 = np.poly1d(A1i[::-1])

A2i = np.array([1, -0.982254, 0.568593, 0.512587, -1.184361, 0.939222, -0.240814])
e2 = np.poly1d(A2i[::-1])

A3i = np.array([0.5, -0.664611, 0.706192, -0.293500, -0.083980, 0.200306, -0.064264])
e3 = np.poly1d(A3i[::-1])

SQRT_PI_8 = np.sqrt(np.pi/8)
def egh_pdf(x, tR=0, sigma=1.0, tau=0.0, scale=1):
    tau_ = abs(tau)
    th = np.arctan2(tau_, sigma)
    M0 = (sigma*SQRT_PI_8 + abs(tau_))*e0(th)
    return egh(x, scale, tR, sigma, tau)/M0

class EghMixture:
    def __init__(self, K, max_iter=100):
        self.K = K
        self.max_iter = max_iter

    def fit(self, X, bins=None):
        assert X.shape[1] == 1
        self.N = X.shape[0]
        self.X = X
        self.X_  = X.flatten()
        self.bins = bins
        self.initilize()

        for step in range(self.max_iter):
            self._e_step()
            self._m_step()

        print('result pi=', self.pi)
        print('result tR=', self.tR)
        print('result sigma=', self.sigma)
        print('result tau=', self.tau)

    def initilize(self):
        # self._initialise_parameters(self.X)
        self.guess_initial_params()

    def calculate_mean_covariance(self, X, prediction):
        """Calculate means and covariance of different
            clusters from k-means prediction

        Parameters:
        ------------
        prediction: cluster labels from k-means
        X: N*d numpy array data points 

        Returns:
        -------------
        intial_means: for E-step of EM algorithm
        intial_cov: for E-step of EM algorithm

        """
        d = X.shape[1]
        labels = np.unique(prediction)
        initial_means = np.zeros((self.K, d))
        initial_cov = np.zeros((self.K, d, d))
        initial_pi = np.zeros(self.K)
        
        counter=0
        for label in labels:
            ids = np.where(prediction == label) # returns indices
            initial_pi[counter] = len(ids[0]) / X.shape[0]
            initial_means[counter,:] = np.mean(X[ids], axis = 0)
            de_meaned = X[ids] - initial_means[counter,:]
            Nk = X[ids].shape[0] # number of data points in current gaussian
            initial_cov[counter,:, :] = np.dot(initial_pi[counter] * de_meaned.T, de_meaned) / Nk
            counter+=1

        # assert np.sum(initial_pi) == 1
        assert abs(np.sum(initial_pi) - 1) < 1e-10
        return (initial_means, initial_cov, initial_pi)

    def _initialise_parameters(self, X):
        """Implement k-means to find starting
            parameter values.
            https://datascience.stackexchange.com/questions/11487/how-do-i-obtain-the-weight-and-variance-of-a-k-means-cluster

        Parameters:
        ------------
        X: numpy array of data points

        Returns:
        ----------
        tuple containing initial means and covariance
        _initial_means: numpy array: (C*d)
        _initial_cov: numpy array: (C,d*d)

        """
        n_clusters = self.K
        kmeans = KMeans(n_clusters= n_clusters, init="k-means++", max_iter=500, algorithm = 'auto')
        fitted = kmeans.fit(X)
        prediction = kmeans.predict(X)
        _initial_means, _initial_cov, _initial_pi = self.calculate_mean_covariance(X, prediction)

        self.pi = _initial_pi
        self.tR = _initial_means.flatten()
        self.sigma = np.sqrt(_initial_cov.flatten())
        self.tau = np.zeros(self.K)
        print('init pi=', self.pi)
        print('init tR=', self.tR)
        print('init sigma=', self.sigma)
        print('init tau=', self.tau)

    def guess_initial_params(self):
        X = self.X
        num_entire_points = X.shape[0]
        kmeans = KMeans(n_clusters= self.K)
        kmeans.fit(X)
        predicted = kmeans.predict(X)
        labels = np.unique(predicted)
        init_pi = []
        init_tR = []
        init_sigma = []

        for label in labels:
            sy = X[predicted == label]

            M1 = np.mean(sy)
            if USE_MODE_AS_TR_INIT:
                # seems risky
                m, c = stats.mode(sy)
                tR = m[0,0]
            else:
                tR = M1

            init_tR.append(tR)
            num_points = sy.shape[0]
            pi = num_points/num_entire_points
            init_pi.append(pi)
            # M2 = pi*np.sum((sy - tR)**2)/num_points     # why pi*?
            M2 = np.sum((sy - M1)**2)/num_points     # why pi*?
            sigma = np.sqrt(M2)
            # print([label], m, c, M2, np.sqrt(M2), sigma)
            init_sigma.append(sigma)

        self.pi = np.array(init_pi)
        self.tR = np.array(init_tR)
        self.sigma = np.array(init_sigma)
        self.tau = np.zeros(self.K)
        print('init pi=', self.pi)
        print('init tR=', self.tR)
        print('init sigma=', self.sigma)
        print('init tau=', self.tau)

        # assert np.sum(initial_pi) == 1
        assert abs(np.sum(self.pi) - 1) < 1e-10

        if False:
            plt.push()
            fig = plt.figure()
            ax = fig.gca()
            ax.hist(X.flatten(), bins=self.bins)
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            for tR in self.tR:
                ax.plot([tR, tR], [ymin, ymax], ':', color='red')
            plt.show()
            plt.pop()

    def _e_step(self):
        self.gamma = np.zeros((self.N, self.K))

        for k in range(self.K):
            # Posterior Distribution using Bayes Rule
            self.gamma[:,k] = self.pi[k] * egh_pdf(self.X_, self.tR[k], self.sigma[k], self.tau[k])

        # normalize across columns to make a valid probability
        # gamma_norm = np.sum(self.gamma, axis=1)[:,np.newaxis]
        gamma_norm = np.sum(self.gamma, axis=1)

        if False:
            print('pi=', self.pi)
            print('gamma_norm=', gamma_norm)
            plt.push()
            fig = plt.figure()
            ax = fig.gca()
            ax.plot(gamma_norm.flatten())
            plt.show()
            plt.pop()

        positive = gamma_norm > 0
        for k in range(self.K):
            self.gamma[positive,k] /= gamma_norm[positive]      # normalize only where positive

    def _m_step(self):

        # responsibilities for each gaussian
        self.pi = np.mean(self.gamma, axis=0)
        self.m1 = np.dot(self.gamma.T, self.X) / np.sum(self.gamma, axis=0)

        W_ = np.sum(self.gamma, axis=0)[:,np.newaxis].T
        # print('GT.shape=', GT.shape)
        # print('GT[:,0:5]=', GT[:,0:5])

        """
        2.1.1 Central Moments and their Distribution Functions
        https://www.iue.tuwien.ac.at/phd/puchner/node17.html#SECTION00811000000000000000
        2.1.2 Probability Weighted Moments and their Distribution Functions
        https://www.iue.tuwien.ac.at/phd/puchner/node18.html

        Central moment - Wikipedia
        https://en.wikipedia.org/wiki/Central_moment
        """

        debug = False

        # M1 = (np.dot(self.gamma.T, self.X) / W_.T).T
        GX = self.gamma * self.X
        M1 = np.sum(GX, axis=0) / W_
        M2 = np.sum(self.gamma * (self.X-M1)**2, axis=0) / W_
        # M3 = np.sum(self.gamma * (self.X-M1)**3, axis=0) / W_

        if debug:
            print('M1=', M1)
            print('M2=', M2)
            print('M3=', M3)

        params = []
        for k in range(self.K):
            params.append(self.solve_params(k, GX[:,k], M1[0,k], M2[0,k]))

        params_ = np.array(params)
        if debug:
            print('params_=', params_)
        self.tR = params_[:,0]
        self.sigma = params_[:,1]
        self.tau = params_[:,2]

    def solve_params(self, k, gx, m1, m2):

        gx_ = gx[gx > 10]

        if False:
            plt.push()
            fig = plt.figure()
            ax = fig.gca()
            ax.hist(gx_, bins=self.bins)
            plt.show()
            plt.pop()

        tR_, _ = stats.mode(gx_)
        tR = tR_[0]
        print([k], ('tR=', tR))

        # last_tR = self.tR[k]
        last_sigma = self.sigma[k]
        last_tau = self.tau[k]

        def equations(p):
            sigma, tau = p
            tau_ = abs(tau)         # better for 2019118_3
            # tau_ = tau            # seems better for some cases such as 20181203, although abs(tau) is mentioned in the paper.
            th = np.arctan2(tau_, sigma)
            return (
                        tR + tau*e1(th) - m1,
                        (sigma**2 + sigma*tau_ + tau**2)*e2(th) - m2,
                        # tau*(3*sigma**2 + 4*sigma*tau_ + 4*tau**2)*e3(th) - m3,
                   )

        sigma, tau = fsolve(equations, (last_sigma, last_tau))

        debug = False

        if debug:
            print([k], (last_tR, '->', tR))
            print([k], (last_sigma, '->', sigma))
            print([k], (last_tau, '->', tau))

        return tR, sigma, tau



import numpy                as np
import matplotlib.pyplot    as plt
import statsmodels.api      as sm

def draw( ax, N, A, B, E ):
    x = np.linspace( 0, 1, N )
    y0  = A + B*x
    y_  = y0 + E * np.random.randn( N )

    stdev = np.std( y_ )
    print( 'std(y_)=', stdev )

    X = sm.add_constant( x )
    result = sm.OLS( y_, X ).fit()
    print( 'result.params=', result.params )
    Va, Vb = np.sqrt( np.diag( result.cov_params() ) )
    print( 'sqrt(cov_params)=', Va, Vb )
    # print( 'result.resid=', result.resid )
    ave_resid = np.sqrt( np.average( result.resid**2 ) )
    print(  'sqrt( average( resid**2 ) )=', ave_resid )


    A_, B_ = result.params
    yr  = A_ + B_ * x

    ax.set_title( 'y=a+b*x; n=%d, a=%.1g, b=%.1g, E=%.1g' % ( N, A, B, E ) )
    ax.set_ylim( 0, 12 )
    if B < 5:
         tx, ty = 0.05, 7
    else:
         tx, ty = 0.5, 1
    ax.text( tx, ty,
                'STDEV=%.3g\nAveraged Error=%.3g\nVar(a_)=%.3g\nVar(b_)=%.3g' %
                ( stdev, ave_resid, Va, Vb )
           )
    ax.plot( x, y_, 'b'  )
    ax.plot( x, y0, 'y' )
    ax.plot( x, yr, 'r' )

fig = plt.figure( figsize=( 18, 9 ) )

axes = [ fig.add_subplot( 4, 4, i+1 ) for i in range(16) ]

A   = 5
E   = 0.2

i = 0
for N in [ 10, 100, 1000, 10000 ]:
    for B in [ 0, 3, -3, 7 ]:
        draw( axes[i], N, A, B, E  )
        i += 1

plt.tight_layout()
plt.show()



"""
    GuinierTools.CompareGdevMeasures.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

def compare_gdev_measures_impl(gdev, Cxr, rg_params, valid_components):
    from importlib import reload
    import GuinierTools.RgCurveUtils
    reload(GuinierTools.RgCurveUtils)
    from .RgCurveUtils import get_connected_curve_info, convert_to_milder_qualities, get_reconstructed_curve
    from Distance.FrobeniusXdiffmax import frobenius_xdiffmax
    print("compare_gdev_measures_impl")
    rg_curve = gdev.rg_curve
    x = rg_curve.x
    y = rg_curve.y
    x_, y_, rgv, qualities, valid_bools = get_connected_curve_info(rg_curve)
    mqualities = convert_to_milder_qualities(qualities)
    size = len(x[valid_bools])
    rrgv = get_reconstructed_curve(size, valid_bools, Cxr, rg_params)

    print("len(x_)=", len(x_))
    print("len(qialities[0])=", len(rg_curve.qualities[0]))
    diff = frobenius_xdiffmax(rgv, rrgv, debug=True)
    print("diff=", diff)

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_title("compare_gdev_measures_impl")
        axt = ax.twinx()
        axt.grid(False)
        ax.plot(x, y, label='Data')
        axt.plot(x_, rgv, label='measured Rg curve', color='gray')
        axt.plot(x_, rrgv, ":", label='reconstructed Rg curve', color='gray')
        axt.plot(x_, qualities*100, label='Quality', color='green', alpha=0.3)
        axt.plot(x_, mqualities*100, label='Milder Quality', color='cyan', alpha=0.3)
        fig.tight_layout()
        plt.show()


"""
    SecTheory.TriporeExclCurve.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import os
import sys
import numpy as np
from matplotlib.gridspec import GridSpec
from scipy.interpolate import UnivariateSpline
from scipy.optimize import minimize, basinhopping

def compute_confpoints_impl(x, rgv, proportions, poresizes, N, T, x0, me, mp):
    import molass_legacy.KekLib.DebugPlot as plt
    from SecTheory.SecCF import gec_dipore_phi, gec_tripore_phi
    from SecTheory.SecPDF import FftInvPdf

    dipore_pdf = FftInvPdf(gec_dipore_phi)
    tripore_pdf = FftInvPdf(gec_tripore_phi)

    proportions_ = proportions/np.sum(proportions)

    np_s = []
    tp_s = []
    for rp, p in zip(poresizes, proportions_):
        rho = rgv/rp
        rho[rho > 1] = 1
        np_ = p*N*(1 - rho)**me
        np_s.append(np_)
        tp_ = T*(1 - rho)**mp
        tp_s.append(tp_)

    confpoints = []
    k = 0

    for np1, tp1, np2, tp2, np3, tp3 in zip(np_s[0], tp_s[0], np_s[1], tp_s[1], np_s[2], tp_s[2]):
        y = tripore_pdf(x, np1, tp1, np2, tp2, np3, tp3, x0)
        j = np.argmax(y)
        confpoints.append((x[j], rgv[k]))
        if False:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("[%d] tripore_pdf debug" % k)
                ax.plot(x, y)
                ax.plot(x[j], y[j], "o", color="red")
                fig.tight_layout()
                plt.show()
        k += 1

    confpoints = np.array(confpoints)
    return confpoints

def create_spline(confpoints, devel=True):
    if devel:
        from importlib import reload
        import SecTheory.ExclCurveUtils
        reload(SecTheory.ExclCurveUtils)
    from SecTheory.ExclCurveUtils import safely_create_spline_from_points
    return safely_create_spline_from_points(confpoints)

def compute_excurve_impl(x, *args):
    confpoints = compute_confpoints_impl(x, *args)
    spline = create_spline(confpoints)
    return spline(x)

def estimate_tripore_exclcurve(fig, ax, axpsd, optimizer, lrf_info, new_params, niter=3):
    import molass_legacy.KekLib.DebugPlot as plt
    from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
    from Experiment.ColumnTypes import get_columnname

    x = optimizer.xr_curve.x
    xr_params = optimizer.separate_params[0]
    # sec_rg_params = optimizer.separate_params[2]
    seccol_params = optimizer.separate_params[7]
    lrf_rg_params = lrf_info.rg_params
    opt_weights = lrf_info.rg_qualities
    print("estimate_exclusion_curve")
    print("\tseccol_params=", seccol_params)
    print("\tlrf_rg_params=", lrf_rg_params)
    print("\topt_weights=", opt_weights)

    Npc, rp, tI, t0, P, m = seccol_params

    trs = xr_params[:,1]

    num_pores = 3

    def objective(p, return_confpoints=False):
        confpoints = compute_confpoints_impl(x, lrf_rg_params, p[0:num_pores], p[num_pores:num_pores*2], *p[num_pores*2:])
        if return_confpoints:
            return confpoints
        return np.sum(opt_weights*(confpoints[:,0] - trs)**2)

    N = P
    T = 1
    x0 = x[0]
    me = 1.5
    mp = 0.1

    rpv = np.linspace(rp*2, rp/2, num_pores)
    init_params = np.concatenate([np.ones(num_pores)/num_pores, rpv, [N, T, x0, me, mp]])
    print("init_params=", init_params)

    last_fun = objective(init_params)

    best_fun = None
    best_params = None

    bounds = [(0, 1)]*num_pores + [(0, 300)]*num_pores + [(0, 2000), (0, 100), (-500, trs[0]), (0, 3), (0, 3)]
    minimizer_kwargs=dict(method='Nelder-Mead', bounds=bounds)

    for k in range(niter):
        print([k], "optmizing ...")
        ret = basinhopping(objective, init_params, minimizer_kwargs=minimizer_kwargs)
        print("ret.success=", ret.success)
        print("ret.x=", ret.x)
        if best_fun is None or ret.fun < best_fun:
            best_fun = ret.fun
            best_params = ret.x
            if best_fun < last_fun:
                init_params = best_params
                last_fun = best_fun

    np.savetxt("best_params.txt", best_params)

    # init_confpoints = objective(init_params, return_confpoints=True)
    opt_confpoints = objective(best_params, return_confpoints=True)

    y = optimizer.xr_curve.y

    in_folder = get_in_folder()
    ax.set_title("Tripore Exclusion Curve on %s with %s" % (in_folder, get_columnname()), fontsize=20)
    ax.set_ylabel("Intensity")

    axt = ax.twinx()
    axt.grid(False)
    axt.set_ylabel("$R_g$")

    axis_info = [fig, (None, ax, None, axt)]    # fig, (ax1, ax2, ax3, axt)
    optimizer.objective_func(new_params, plot=True, axis_info=axis_info)

    opt_proportions = best_params[0:num_pores]
    opt_poresizes = best_params[num_pores:num_pores*2]
    opt_extraparams = best_params[num_pores*2:]
    opt_x0 = opt_extraparams[2]

    maxrg = np.max(lrf_rg_params)*1.1
    minrg = 15

    rgv = np.arange(maxrg, minrg, -1)
    confpoints = compute_confpoints_impl(x, rgv, opt_proportions, opt_poresizes, *opt_extraparams)

    if False:
        cwd = os.getcwd()
        print("cwd=", cwd)

        np.savetxt("confpoints.txt", confpoints)

        with plt.Dp():
            fig_, ax_ = plt.subplots()
            ax_.set_title("confpoints debug")
            ax_.plot(*confpoints.T, "o")
            fig_.tight_layout()
            plt.show()

    spline = create_spline(confpoints)
    x_ = x[x > opt_x0]
    r_ = spline(x_)
    r__ = r_[r_ > minrg]
    x__ = x_[r_ > minrg]

    axt.plot(x__, r__, color="cyan", markersize=1, lw=3, label="tripore exclusion curve")

    # axt.plot(*init_confpoints.T, "o", color="blue")
    axt.plot(*opt_confpoints.T, "o", color="red", label="elution points")
    axt.set_ylim(0, maxrg)
    axt.legend(bbox_to_anchor=(0.7, 1), loc="upper center")

    opt_proportions = opt_proportions/np.sum(opt_proportions)

    axpsd.set_title("Poresize Distribution", fontsize=20)
    axpsd.bar(opt_poresizes, opt_proportions, width=10)
    axpsd.set_xlim(0, 300)

def demo():
    import molass_legacy.KekLib.DebugPlot as plt

    x = np.arange(200)
    rp = 120

    rgv = np.flip(np.linspace(0, rp-1, 50))

    x0 = 50
    N = 171
    T = 0.63
    me = 1.5
    mp = 0.1

    proportions3 = np.ones(3)/3
    poresizes3_1 = np.array([rp, rp, rp])
    poresizes3_2 = np.array([rp*2, rp, rp/2])

    y3 = compute_excurve_impl(x, rgv, proportions3, poresizes3_1, N, T, x0, me, mp)
    y4 = compute_excurve_impl(x, rgv, proportions3, poresizes3_2, N, T, x0, me, mp)

    valid = np.logical_and(x > x0, y3 > 0)
    x_ = x[valid]
    y3_ = y3[valid]
    y4_ = y4[valid]

    with plt.Dp():
        fig, ax = plt.subplots()

        ax.plot(x_, y3_, label=str(poresizes3_1))
        ax.plot(x_, y4_, label=str(poresizes3_2))

        ax.legend()
        fig.tight_layout()
        plt.show()

if __name__ == '__main__':
    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.dirname(this_dir)
    sys.path.append(lib_dir)

    import seaborn
    seaborn.set()
    import molass_legacy.KekLib

    demo()



# coding: utf-8
"""
    RdAnim.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.SciPyCookbook import smooth
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from tkinter import ttk
from SvdDenoise import get_denoised_data
from DataUtils import get_in_folder
from SerialAtsasTools import AlmergeExecutor
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.KekLib.BasicUtils import clear_dirs_with_retry

WIDTH = 10

class RdAnim(Dialog):
    def __init__( self, parent, sd, rank=2, sythesized=False):
        self.sd = sd
        self.j0 = self.sd.xr_j0
        assert rank in [1, 2]
        self.rank = rank
        self.sythesized = sythesized
        self.almerge = AlmergeExecutor()
        self.prepare_curves()
        self.mode = Tk.IntVar()
        self.mode.set(1)
        self.hw = WIDTH//2
        Dialog.__init__( self, parent, "RD Animation", visible=False )

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()

        fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(21, 11))
        self.fig = fig
        self.axes = axes
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        fig.tight_layout()
        fig.subplots_adjust(top=0.9, bottom=0.05, left=0.05, right=0.95, wspace=0.2)
        self.draw()
        self.mpl_canvas.draw()

    def buttonbox(self):
        frame = Tk.Frame(self)
        frame.pack(fill=Tk.X)

        box = Tk.Frame(frame)
        box.pack(side=Tk.LEFT, padx=200)
        panel = Tk.Frame(frame)
        panel.pack(side=Tk.RIGHT, padx=200)

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.build_panel(panel)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def build_panel(self, frame):
        w = Tk.Checkbutton(frame, text="animate", variable=self.mode)
        w.pack(side=Tk.LEFT, padx=20)

        f, t = self.get_range()
        self.pos = Tk.IntVar()
        self.pos.set(f+self.hw)
        scale = ttk.Scale(
            frame,
            variable=self.pos,
            orient=Tk.HORIZONTAL, 
            length=200,
            from_=f,
            to=t-WIDTH-1)
        scale.pack(side=Tk.LEFT, padx=20)

        spinbox = ttk.Spinbox(frame, textvariable=self.pos, format="%3.0f",
                                  from_=f, to=t-WIDTH-1, increment=1,
                                  justify=Tk.CENTER, width=6)
        spinbox.pack(side=Tk.LEFT, padx=20)
        self.mode_dependents = [scale, spinbox]
        self.mode_tracer()

        self.mode.trace("w", self.mode_tracer)
        self.pos.trace("w", self.position_changed)

    def mode_tracer(self, *args):
        mode = self.mode.get()
        state = Tk.DISABLED if mode else Tk.NORMAL
        for w in self.mode_dependents:
            w.config(state=state)

    def position_changed(self, *args):
        mode = self.mode.get()
        if mode:
            return

        f, t = [p + self.hw for p in self.get_range()]
        try:
            value = self.pos.get()
            i = int(round(value))
            if f <= i and i < t-WIDTH:
                self.pos.set(i)
        except:
            pass

    def prepare_curves(self):
        if self.sythesized:
            from Rank.Synthesized import synthesized_data
            self.Rg = sd.pre_recog.get_rg()

        M, E, qv, ecurve = self.sd.get_xr_data_separate_ly()
        self.qv = qv
        self.ecurve = ecurve
        self.anim_range = self.ecurve.get_ranges_by_ratio(0.1)[0]
        print('anim_range=', self.anim_range)
        print('j0=', self.j0)
        f, t = self.get_range()

        y = ecurve.y
        c = y/np.max(y)

        self.scurve_ys_list = [[], []]
        self.switch_points = []
        if self.rank == 2:
            self.bq_ys_list = [[], []]
            self.bq_lims = []
            bq_slice = slice(0, int(len(qv)*0.1))
        for k in range(f, t-WIDTH):
            j_ = slice(k, k+WIDTH)
            Mj = M[:,j_]
            c_ = c[j_]
            sp = self.get_almerge_switch_point(Mj, E[:,j_], c_)
            print([k], 'sp=', sp)
            self.switch_points.append(sp)

            if self.rank == 1:
                C_ = np.array([c_])
                Mj_ = get_denoised_data(Mj, rank=self.rank)
            else:
                C_ = np.array([c_, c_**2])
                if self.sythesized:
                    Mj_, _ = synthesized_data(self.qv, Mj, self.Rg)
                else:
                    Mj_ = get_denoised_data(Mj, rank=self.rank)

            for m, M_ in enumerate([Mj, Mj_]):
                scurve_ys = self.scurve_ys_list[m]
                P_ = M_ @ np.linalg.pinv(C_)
                scurve_ys.append(P_[:,0])
                if self.rank == 2:
                    bq_ys = self.bq_ys_list[m]
                    bq = P_[:,1]
                    bq_ys.append(bq)
                    self.bq_lims.append((np.min(bq[bq_slice]), np.max(bq[bq_slice])))

        self.noises_list = []
        for scurve_ys in self.scurve_ys_list:
            noises = []
            for k, y in enumerate(scurve_ys):
                j = f+k
                y_ = y/E[:,j]
                sy = smooth(y_)
                n = np.linalg.norm(y_ - sy)
                noises.append(n)
            self.noises_list.append(np.array(noises))

    def get_almerge_switch_point(self, Mj, Ej, c):
        temp_folder = get_setting('temp_folder')
        qv = self.qv
        files = []
        for k in range(Mj.shape[1]):
            data = np.array([qv, Mj[:,k], Ej[:,k]]).T
            file = os.path.join(temp_folder, 'data-%03d.dat' % k)
            np.savetxt(file, data)
            files.append(file)
        indeces = np.arange(Mj.shape[1], dtype=int)
        out_file = os.path.join(temp_folder, 'extrapolated.dat')
        result = self.almerge.execute(c, files, indeces, out_file)
        clear_dirs_with_retry([temp_folder])
        return result.overlap_from_max

    def get_range(self):
        f = self.anim_range[0]
        t = self.anim_range[2] + 1
        return f, t

    def draw(self):
        fig = self.fig
        ax1, ax2, ax3 = self.axes[0,:]
        ax4, ax5, ax6 = self.axes[1,:]

        fig.suptitle("Range Dependency Animation for %s with Rank(:, %d)" % (get_in_folder(), self.rank), fontsize=30)
        ax1.set_title("Elution Curve", fontsize=24)
        ax2.set_title("LRF'ed Scattering Curve", fontsize=24)
        ax3.set_title("Noise Levels", fontsize=24)

        for ax in [ax2, ax5]:
            ax.set_yscale('log')
        ecurve = self.ecurve

        x = ecurve.x + self.j0
        y = ecurve.y
        switch_points = self.switch_points

        f, t = self.get_range()
        slice_ = slice(f, f+WIDTH)
        dcurves = []
        for ax in [ax1, ax4]:
            ax.plot(x, y, color='orange')
            dcurve, = ax.plot(x[slice_], y[slice_], color='red', linewidth=5, alpha=0.5)
            dcurves.append(dcurve)

        qv = self.qv
        if self.rank == 2:
            axt_bqs = []
            bcurves = []
        scurves = []
        m = 0
        for ax, ys in zip([ax2, ax5], self.scurve_ys_list):
            if self.rank == 2:
                axt = ax.twinx()
                axt.grid(False)
                axt_bqs.append(axt)
            scurve, = ax.plot(qv, ys[0], label="Aq")
            scurves.append(scurve)
            if self.rank == 2:
                bq_ys = self.bq_ys_list[m]
                bcurve, = axt.plot(qv, bq_ys[0], color='pink', label='Bq')
                bcurves.append(bcurve)
            m += 1

        axts = []
        points = []
        for ax, noises in zip([ax3, ax6], self.noises_list):
            ax.plot(x, y, ':', color='orange')
            axt = ax.twinx()
            ax.grid(False)
            axts.append(axt)
            nj_ = slice(f, t-WIDTH)
            axt.plot(x[nj_]+self.hw, noises)
            point, = axt.plot(x[f]+self.hw, noises[0], 'o', color='red')
            points.append(point)

        def set_wider_ylim(axes, ylim=None):
            if ylim is None:
                ylims = np.array([ax.get_ylim() for ax in axes])
                ymin = np.min(ylims[:,0])
                ymax = np.max(ylims[:,1])
            else:
                ymin, ymax = ylim
            for ax in axes:
                ax.set_ylim(ymin, ymax)

        set_wider_ylim([ax2, ax5], ylim=[1e-5, 1])
        if self.rank == 2:
            bq_lims = np.array(self.bq_lims)
            ymin = np.percentile(bq_lims[:,0], 20)
            ymax = np.percentile(bq_lims[:,1], 80)
            set_wider_ylim(axt_bqs, ylim=[ymin, ymax])
        set_wider_ylim(axts)

        sp = qv[switch_points[0]]
        ymin2, ymax2 = ax2.get_ylim()
        switch_line, = ax2.plot([sp, sp], [ymin2, ymax2], color='red', label='almerge boundary')

        for ax in [ax2, ax5]:
            ax.legend()
        if self.rank == 2:
            for ax in axt_bqs:
                ax.legend(bbox_to_anchor=(1, 0.85), loc='upper right')

        artists = dcurves + scurves + points + [switch_line]
        if self.rank == 2:
            artists += bcurves

        def get_dcurve_data(i):
            f_ = f+i
            slice_ = slice(f_, f_+WIDTH)
            return x[slice_], y[slice_]

        def init():
            return artists

        def animate(i):
            mode = self.mode.get()
            if mode:
                self.pos.set(f+i+self.hw)
            else:
                i = self.pos.get() - self.hw - f
            xy = get_dcurve_data(i)
            for dcurve in dcurves:
                dcurve.set_data(*xy)
            for scurve, ys in zip(scurves, self.scurve_ys_list):
                scurve.set_data(qv, ys[i])
            if self.rank == 2:
                for bcurve, bq_ys in zip(bcurves, self.bq_ys_list):
                    bcurve.set_data(qv, bq_ys[i])
            for point, noises in zip(points, self.noises_list):
                point.set_data(x[f+i]+self.hw, noises[i])

            sp = qv[switch_points[i]]
            switch_line.set_data([sp, sp], [ymin2, ymax2])
            return artists

        num_frames = t - WIDTH - f
        self.anim = animation.FuncAnimation(self.fig, animate, init_func=init,
                                       frames=num_frames, interval=500, blit=True)



"""
    OurManim.py

    Copyright (c) 2019-2024, SAXS Team, KEK-PF
"""
from time import sleep
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import rcParams
from matplotlib import animation
from matplotlib.patches import Polygon, Circle
from molass_legacy.KekLib.OurMatplotlib import mpl_1_5_backward_compatible_init

def latex_init():
    rcParams['text.usetex'] = True
    rcParams['text.latex.preamble'] = r'\usepackage{amsmath}'

def manim_init():
    plt.close()
    plt.style.use('dark_background')
    latex_init()

def use_default_style():
    plt.close()
    plt.style.use('default')
    mpl_1_5_backward_compatible_init()
    latex_init()

def rotation(theta):
    return np.array([[np.cos(theta), -np.sin(theta)],
                  [np.sin(theta),  np.cos(theta)]])

def angle(v0, v1):
    """
    Angles between two n-dimensional vectors in Python
    https://stackoverflow.com/questions/2827393/angles-between-two-n-dimensional-vectors-in-python
    """
    return np.math.atan2(np.linalg.det([v0,v1]),np.dot(v0,v1))

def rotate(theta, vertices):
    R = rotation(theta)
    return list(np.dot(R, np.array(vertices).T).T)

class Animation:
    def __init__(self, start, stop, on_stop=None):
        self.start = start
        self.stop = stop
        self.on_stop = on_stop
        self.count = stop - start
        self.slice_ = slice(start, stop)
        self.anim_objects = []
        self.artists = []

    def close(self):
        print('close: ', self.stop)
        if self.on_stop is not None:
            print('on_stop: ', self.stop)
            self.on_stop()
        self.set_visible(False)
        # return self.artists + self.close_artists
        return self.artists

    def append(self, anim_object):
        anim_object.divide(self.count)
        self.anim_objects.append(anim_object)
        self.artists += anim_object.get_artists()

    def get_artists(self):
        return self.artists

    def animate(self, i):
        assert i >= self.start
        # print([i], "- animate")
        for aobj in self.anim_objects:
            if i < self.stop:
                aobj.animate(i - self.start)
            else:
                if i == self.stop:
                    self.close()
                break
        return self.artists

    def set_visible(self, torf):
        for a in self.artists:
            a.set_visible(torf)

class Collection:
    def __init__(self, num_frames, ret_artists=[], on_reset=None):
        self.num_frames = num_frames
        self.ret_artists = ret_artists
        self.on_reset = on_reset
        self.anim_dict = {}
        """
        indeces = {
                    0 : [a0, a1],
                    1 : [a0, a1],
                    2 : [a0,    a2, ...],
                    3 : [       a2, ...],
                    ...
                  }
        """
        self.last_anims = None

    def append(self, anim_object):
        for i in range(anim_object.slice_.start, anim_object.slice_.stop):
            anims = self.anim_dict.get(i)
            if anims is None:
                self.anim_dict[i] = anims = []
            anims.append(anim_object)
        anim_object.set_visible(False)

    def make_animation(self, fig):
        anim = animation.FuncAnimation(fig, self.animate, frames=self.num_frames, blit=True, init_func=self.reset)
        return anim

    def reset(self):
        print('reset ------------------------------')

        if self.on_reset is not None:
            self.on_reset()

        hidden_artists = []
        if self.last_anims is not None:
            for anim in self.last_anims:
                anim.set_visible(False)
                hidden_artists += anim.get_artists()

        init_artists = []
        init_anims = self.anim_dict.get(0)
        if init_anims is not None:
            for anim in init_anims:
                anim.set_visible(True)
                init_artists += anim.get_artists()

        return self.ret_artists + hidden_artists + init_artists

    def animate(self, i):
        # sleep(1)
        # print([i], "animate")
        anims = self.anim_dict.get(i)
        if anims is None:
            return self.ret_artists

        active_artists = []
        for anim in anims:
            anim.set_visible(True)
            active_artists += anim.animate(i)

        close_artists = []
        if self.last_anims is not None:
            for anim in self.last_anims:
                if i == anim.slice_.stop:
                    close_artists += anim.close()

        self.last_anims = anims
        return self.ret_artists + active_artists + close_artists

class TextGroup:
    def __init__(self, ax, texts, positions, **kwargs):
        self.ax = ax
        self.texts = texts
        self.init_positions = np.array(positions)
        self.artists = []
        for t, p in zip(texts, positions):
            self.artists.append(ax.text(*p, t, **kwargs))

    def get_artists(self):
        return self.artists

    def set_target_positions(self, positions):
        self.target_positions = np.array(positions)

    def divide(self, count):
        self.deltas = (self.target_positions - self.init_positions)/count

    def set_visible(self, torf):
        for a in self.artists:
            a.set_visible(torf)

    def reset_positions(self):
        for k, a in enumerate(self.artists):
            a.set_position(self.init_positions[k,:])

    def animate(self, i):
        # print([i], "-- animate")
        new_positions = self.init_positions + (i+1)*self.deltas
        for k, a in enumerate(self.artists ):
            a.set_position(new_positions[k,:])

class Parallelogram:
    def __init__(self, ax, vertices, motion='linear', **kwargs):
        self.init_vertices = vertices
        self.polygon = Polygon(vertices, **kwargs)
        self.motion = motion
        ax.add_patch(self.polygon)

    def set_target_vertices(self, vertices):
        self.target_vertices = np.array(vertices)

    def divide(self, count):
        if self.motion == 'rotation':
            v0 = self.init_vertices[1]
            v1 = self.target_vertices[1]
            a = angle(v0, v1)
            print('angle=', np.degrees(a))
            self.theta = a/count
        else:
            self.deltas = (self.target_vertices - self.init_vertices)/count

    def get_artists(self):
        return [self.polygon]

    def animate(self, i):
        if self.motion == 'rotation':
            theta = (i+1)*self.theta
            new_vertices = rotate(theta, self.init_vertices)
        else:
            new_vertices = self.init_vertices + (i+1)*self.deltas
        self.polygon.set_xy(new_vertices)

    def make_reversed(self, ax, **kwargs):
        para = Parallelogram(ax, self.target_vertices, motion=self.motion, **kwargs)
        para.set_target_vertices(self.init_vertices)
        return para

class Arrow:
    """
    this class is incomplete,
    because the head shape won't be consistent with the move direction.
    """
    def __init__(self, ax, pos, dpos, **kwargs):
        self.init_position = (pos, dpos)
        self.arrow = ax.arrow(*pos, *dpos, **kwargs)
        ax.add_patch(self.arrow)
        self.ax = ax
        self.init_vertices = self.arrow.get_xy()
        self.init_kwartgs = kwargs

    def set_target_position(self, pos, dpos):
        self.target_position = (pos, dpos)
        temp_arrow = self.ax.arrow(*pos, *dpos, visible=False, **self.init_kwartgs)
        self.target_vertices = temp_arrow.get_xy()

    def divide(self, count):
        self.deltas = (self.target_vertices - self.init_vertices)/count

    def get_artists(self):
        return [self.arrow]

    def animate(self, i):
        new_vertices = self.init_vertices + (i+1)*self.deltas
        self.arrow.set_xy(new_vertices)

class Circles:
    def __init__(self, ax, pos, **kwargs):
        self.init_positions = np.array(pos)
        self.circles = [Circle(xy=p, **kwargs) for p in pos]
        for c in self.circles:
            ax.add_patch(c)

    def set_target_positions(self, pos):
        self.target_positions = np.array(pos)

    def divide(self, count):
        self.deltas = (self.target_positions - self.init_positions)/count

    def get_artists(self):
        return self.circles

    def animate(self, i):
        new_positions = self.init_positions + (i+1)*self.deltas
        for c, p in zip(self.circles, new_positions):
            c.set_center(p)



"""
    HdcTheory.LaminarFlow.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
import seaborn as sns
import matplotlib as mpl
# print(mpl.style.available)
# sns.set_theme()
mpl.style.use('dark_background')

def demo():

    tv = np.linspace(0, 0.1, 10)

    L = 30
    R = 1
    margin = 0.1
    y = np.linspace(margin, R*2 - margin, 100)
    mv = np.ones(len(y))*margin
    # rv = np.max([mv, np.abs(y - R)], axis=0)
    rv = np.abs(y - R)
    lv = rv/R
    u = 1 - lv**2

    fig, ax = plt.subplots(figsize=(10,2))
    ax.set_title("Animation of Laminar Flow in a Pipe")
    ax.set_axis_off()

    curves = []
    for t in tv:
        x = u*L*t**2/2
        curve, = ax.plot(x, y, "o", markersize=1)
        curves.append(curve)

    ax.set_xlim(0, 10)

    dt = 0.1
    def update(i):
        elapse_time = i*0.02 
        for k, t in enumerate(elapse_time + tv):
            x = u*L*t**2/2
            # q, r = divmod(x, 100)
            curves[k].set_xdata(x)
        return curves

    def init():
        return update(0)

    num_frames = 100
    anim = FuncAnimation(fig, update, init_func=init,
                               frames=num_frames, interval=100, blit=True)
    fig.tight_layout()
    # anim.save("laminar-flow.gif")
    plt.show()


if __name__ == "__main__":
    import sys
    sys.path.append("../lib")    
    # demo()
    demo()



# coding: utf-8
"""

    PeakMappingRefiner.py

    Copyright (c) 2019, SAXS Team, KEK-PF

"""
import copy
import numpy            as np
import logging
from scipy.stats        import pearsonr, linregress
from scipy.interpolate  import UnivariateSpline
from itertools          import combinations
import OurStatsModels   as sm
from molass_legacy.SerialAnalyzer.ElutionCurve       import ElutionCurve
from molass_legacy.Elution.CurveUtils  import simple_plot
from molass_legacy.KekLib.SciPyCookbook      import smooth
from molass_legacy.KekLib.BasicUtils         import Struct
from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
import molass_legacy.KekLib.DebugPlot        as plt

NUM_ROOTS_TO_AVOID_NOISE    = 5     # dupicate

class PeakMappingRefiner:
    def __init__(self):
        self.debug = False

    def find_counter_part( self, start, stop, top_x1, peak_rec, A_, B_, curve2 ):
        x = curve2.x
        y = curve2.y
        mapped_top_x = A_ * top_x1 + B_
        slice_ = slice(start, stop)
        x_ = x[slice_]
        y_ = y[slice_]

        if len(x_) < 3:
            raise RuntimeError("Invalid slice: " + str((start, stop)))

        slope, intercept = linregress( x_, y_)[0:2]
        baseline = slope * x_ + intercept
        ry_ = y_ - baseline

        try:
            sry_ = smooth(ry_, window_len=20)
        except ValueError:
            raise RuntimeError("smooth failed. for len(ry_)=%d." % len(ry_))

        neighbor_spline = UnivariateSpline(x_, sry_, s=0, ext=3)
        d1 = neighbor_spline.derivative(1)
        d1y = d1(x_)
        d1_spline = UnivariateSpline(x_, d1y, s=0, ext=3)
        d1_roots = d1_spline.roots()

        if self.debug:
            print('len(d1_roots)=', len(d1_roots) )

        # assert len(d1_roots) < NUM_ROOTS_TO_AVOID_NOISE

        d2 = neighbor_spline.derivative(2)
        top_candidates = d1_roots[ d2(d1_roots) < 0 ]

        assert len(top_candidates) > 0, "len(top_candidates) == 0"

        dist = np.abs(top_candidates - mapped_top_x)
        n = np.argmin(dist)
        top_x = top_x_c = top_candidates[n]

        if curve2.spline(mapped_top_x) > curve2.spline(top_x):
            top_x = mapped_top_x

        # assert curve2.locally_tall_enough(top_x)      # not appropriate for SUB_TRN1

        prev_boundaries = d1_roots[ np.logical_and(d1_roots < top_x, d2(d1_roots) > 0) ]
        if len(prev_boundaries) > 0:
            prev_boundary = int(prev_boundaries[-1] + 0.5)
        else:
            # TODO:
            prev_boundary = start

        next_boundaries = d1_roots[ np.logical_and(d1_roots > top_x, d2(d1_roots) > 0) ]
        if len(next_boundaries) > 0:
            next_boundary = int(next_boundaries[0] + 0.5)
        else:
            # TODO
            next_boundary = stop

        r_noise = np.average(np.abs(ry_ - sry_))
        boundaries = [prev_boundary, next_boundary]
        r_height = np.max(neighbor_spline([top_x, top_x_c])) - np.max(neighbor_spline(boundaries))
        # print('r_noise/r_height=', r_noise/r_height)
        r_noise_ratio = r_noise/r_height
        # assert r_noise_ratio < 0.5, "r_noise_ratio(%.3g) >= 0.5" % (r_noise_ratio)

        top_x_i = int(top_x + 0.5)
        range_info = [prev_boundary, top_x_i, next_boundary]

        if False:
            plt.push()
            fig, axes = plt.subplots( figsize=(21, 7), nrows=1, ncols=3 )
            fig.suptitle("find_counter_part debug")
            ax1, ax2, ax3 = axes
            ax1.set_title("Whole View")
            ax2.set_title("Neighbor View")
            ax3.set_title("Transformed View")
            ax1.plot(x, y)
            for k, ax in enumerate(axes):
                if k < 2:
                    ax.plot(x_, y_, label='neighborhood')
                    ax.plot(x_, baseline, ':', color='red', label='baseline')
                    ax.plot(mapped_top_x, curve2.spline(mapped_top_x), 'o', color='yellow', label='mapped peak top')
                    ax.plot(boundaries, curve2.spline(boundaries), 'o', color='green', label='new boundaries')
                    ax.plot(top_candidates, curve2.spline(top_candidates), 'o', color='pink', markersize=8, label='top_candidates')
                    ax.plot(top_x, curve2.spline(top_x), 'o', color='red', label='new peak top')
                else:
                    ax.plot(x_, ry_, label='neighborhood')
                    ax.plot(x_, sry_, label='smoothed neighborhood')
                    ax.plot(d1_roots, neighbor_spline(d1_roots), 'o', color='cyan', markersize=10, label='d1_roots')
                    ax.plot(boundaries, neighbor_spline(boundaries), 'o', color='green', label='new boundaries')
                    ax.plot(top_candidates, neighbor_spline(top_candidates), 'o', color='pink', label='top_candidates')
                    ax.plot(top_x, neighbor_spline(top_x), 'o', color='red', label='new peak top')

                ax.legend()
            fig.tight_layout()
            fig.subplots_adjust(top=0.9)
            plt.show()
            plt.pop()

        # assert len(top_candidates) == 1

        peak_rec = [int(A_*j + B_ + 0.5) for j in peak_rec]

        bottom_info = Struct(prev_bottom=prev_boundary, next_bottom=next_boundary )

        return Struct(top_x=top_x, peak_rec=peak_rec, bottom_info=bottom_info, boundary_score=0, range_info=range_info)

    def maps_well_linearly(self, curve1, curve2, A, B):
        if len(curve1.peak_info) == 1:
            return True

        y = curve1.peak_top_x + curve1.boundaries
        x = curve2.peak_top_x + curve2.boundaries

        correlation = pearsonr(x, y)[0]
        print('correlation=', correlation)  # correlation = 

        if correlation > ADEQUACY_CORRELATION:
            return True

        if len(curve1.boundaries) < 2:
            # 20181203
            # better check if it is easily done
            return correlation > RISKY_CORRELATION_LIMIT

        # in cases like 20180206
        for k in range(len(curve1.boundaries)):
            boundaries1 = copy.deepcopy(curve1.boundaries)
            boundaries2 = copy.deepcopy(curve2.boundaries)
            boundaries1.pop(k)
            boundaries2.pop(k)
            y = curve1.peak_top_x + boundaries1
            x = curve2.peak_top_x + boundaries2

            correlation = pearsonr(x, y)[0]
            print([k], 'correlation=', correlation)  # correlation = 
            if correlation > ADEQUACY_CORRELATION:
                b1 = curve1.boundaries[k]
                mb = int(A*curve2.boundaries[k] + B + 0.5)
                curve1.boundaries[k] = mb
                self.logger.info("moved %dth bottom from %d to %d." % (k, b1, mb))
                return True

        return False

    def refine_mapping_params(self, curve1, curve2, max_params, max_correl, max_simil, index, reversed_):
        # assert False
        B, A = max_params

        top_x1 = curve1.peak_top_x
        top_x2 = curve2.peak_top_x
        pk_info1 = curve1.peak_info
        pk_info2 = curve2.peak_info

        x = []
        y = []
        w = []

        for m in range(len(pk_info1)):
            y.append(top_x1[m])
            x.append(top_x2[m])
            w.append(8)
            rec1 = pk_info1[m]
            rec2 = pk_info2[m]
            for k in [0,2]:
                y.append(rec1[k])
                x.append(rec2[k])
                w.append(1)

        X   = sm.add_constant(x)
        mod = sm.WLS( y, X, weights=w )
        res = mod.fit()
        rB, rA = res.params

        correl, simil = self.compute_mapped_curve_simimarity(res.params, x, y, w, index, curve1, curve2, reversed_ )

        if False:
            from CallStack import CallStack
            cstack = CallStack()
            print('refine_mapping_params call stack=', cstack)
            plt.push()
            fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14, 6))
            ax1, ax2 = axes

            ax1.plot(curve1.x, curve1.y)
            x_ = A * curve2.x + B
            y_ = curve2.y * curve1.max_y/curve2.max_y
            ax1.plot(x_, y_)

            ax2.plot(curve1.x, curve1.y)
            x_ = rA * curve2.x + rB
            y_ = curve2.y * curve1.max_y/curve2.max_y
            ax2.plot(x_, y_)

            fig.tight_layout()
            plt.show()
            plt.pop()

        assert correl + simil > max_correl + max_simil

        return rA, rB

    def peak_tops_correct(self, curve1, m, peak_top_x1, peak_info1, curve2, k, peak_top_x2, peak_info2, A, B):
        scale = curve2.height/curve1.height
        peak_height1 = curve1.spline(peak_top_x1) * scale
        peak_height2 = curve2.spline(peak_top_x2)
        if peak_height1 > peak_height2:
            A_  = 1/A
            B_  = -B/A
            peak_top_x2 = A_ * peak_top_x1 + B_
            peak_info2  = [ int(A_ * p + B_ + 0.5 ) for p in peak_info1 ]
            if m > 0 and m - 1 < len(curve1.boundaries):
                boundary1 = curve1.boundaries[m - 1]
                boundary2 = A_ * boundary1 + B_
            else:
                assert False
        else:
            peak_top_x1 = A * peak_top_x2 + B
            peak_info1  = [ int(A * p + B + 0.5 ) for p in peak_info2 ]
            if k > 0 and k - 1 < len(curve2.boundaries):
                boundary2 = curve2.boundaries[k - 1]
                boundary1 = A * boundary2 + B
            else:
                assert False
        return Struct(
                    peak_top_x1=peak_top_x1, 
                    peak_info1=peak_info1, 
                    prev_boundary1=boundary1, 
                    peak_top_x2=peak_top_x2, 
                    peak_info2=peak_info2, 
                    prev_boundary2=boundary2
                    )



"""
    Optimizer.Devel.CodePatch.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

def code_patch(optimizer, work_params, current_folder):
    from importlib import reload
    import Optimizer.FuncReloadUtils
    reload(Optimizer.FuncReloadUtils)
    from molass_legacy.Optimizer.FuncReloadUtils import reload_optimizer

    temp_optimizer = reload_optimizer(optimizer, init_params=work_params)

    restart_params_file = os.path.join(current_folder, "restart_params.txt")

    def save_params():
        np.savetxt(restart_params_file, work_params)

    def load_params():
        temp_params = np.loadtxt(restart_params_file)
        print("load_params: temp_params=", temp_params)
        for i, val in enumerate(temp_params):
            work_params[i] = val

    extra_button_specs = [
        ("save_params", save_params),
        ("load_params", load_params),
    ]

    with plt.Dp(button_spec=["OK", "Cancel"], extra_button_specs=extra_button_specs):
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,5))
        fig.suptitle("Code Patch Preview")

        axt = ax2.twinx()
        axt.grid(False)
        axes = (ax1, ax2, ax3, axt)
        temp_optimizer.objective_func(work_params, plot=True, axis_info=(fig, axes))

        fig.tight_layout()
        plt.show()


# coding: utf-8
"""

    ファイル名：   OurMatplotlib.py

    処理内容：

       matplotlib 関連の追加クラス

    Copyright (c) 2016-2024, Masatsuyo Takahashi, KEK-PF

"""
import os
import sys
import re
import numpy as np
import pylab as pl
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib import colors
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk as NavigationToolbar
from ControlKeyState    import get_shift_key_state

default_colors = None
hex_colors = None

def mpl_1_5_backward_compatible_init( seaborn=True ):
    # for matplotlib 1.5 backward com;patibility
    # to be removed soon
    mpl.rcParams['figure.figsize'] = [8.0, 6.0]
    mpl.rcParams['figure.dpi'] = 80
    mpl.rcParams['savefig.dpi'] = 100

    mpl.rcParams['font.size'] = 12
    mpl.rcParams['legend.fontsize'] = 'large'
    mpl.rcParams['figure.titlesize'] = 'medium'

    if seaborn:
        import seaborn
        seaborn.set_theme()

    # TODO: unify these initializations with get_default_colors, get_hex_color
    global default_colors, hex_colors
    default_colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
    hex_colors = [ colors.to_hex(c) for c in default_colors ]

def mpl_font_init():
    """
    learned at
    https://stackoverflow.com/questions/35593345/how-can-i-check-the-default-font-of-matplotlib-in-the-python-shell
    """
    mpl.rcParams['font.family'] = ['serif'] # default is sans-serif
    mpl.rcParams['font.serif'] = [
               'Lucida Sans Unicode',
               'cambria',
               'MS Reference Sans Serif',
               'serif',
               ]

def convert_to_fixed_font_latex( text ):
    """
        This function is a workaround to use a fixed font
        only for lengeds.
    """
    # \m causes re.error: bad escape \m at position 1 in python 3.7
    text = re.sub( r'(\S+)', r'$/mathtt{\g<1>}$', text ).replace('/', '\\')
    return text

def get_default_colors():
    # be sure that this call comes after the call to seaborn.set()
    global default_colors
    if default_colors is None:
        # https://stackoverflow.com/questions/42086276/get-default-line-colour-cycle
        # note that this list of colors differs after the call to seaborn.set()
        default_colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
    return default_colors

def get_color(k):
    if default_colors is None:
        get_default_colors()
    return default_colors[k%len(default_colors)]

def get_hex_color(k):
    global hex_colors
    if hex_colors is None:
        hex_colors = [ colors.to_hex(c) for c in default_colors ]
    return hex_colors[k%len(hex_colors)]

def cc(arg):
    return colors.to_rgba(arg, alpha=0.6)

def get_facecolors(num_colors, start=0):
    return [cc('C%d' % ((c+start)%8)) for c in range(num_colors)]

class CoordinateFormatter:
    def __init__( self, numrows, numcols, im_array_list, value_shift=0 ):
        self.numrows = numrows
        self.numcols = numcols
        self.im_array_list = im_array_list
        self.value_shift = value_shift
        self.ic = None
        self.ir = None

    def __call__( self, x, y ):
        self.ic = int(x + 0.5)
        self.ir = int(y + 0.5)
        if self.ic >= 0 and self.ic < self.numcols and self.ir >= 0 and self.ir < self.numrows:
            intensities = []
            for im_array in self.im_array_list:
                intensities.append( '%d' % ( im_array[self.ir, self.ic] -self.value_shift ) )
            return 'x=%d, y=%d, intensity=[ %s ]' % (self.ic, self.ir, ','.join( intensities ) )
        else:
            return 'x=%d, y=%d' % ( self.ic, self.ir )

class DataCursor(object):
    text_template = 'x: %d\ny: %d\ni: %d'
    x, y = None, None
    xoffset, yoffset = -20, 20

    def __init__(self, ax_list, action, value_shift=0 ):

        self.ax_list    = ax_list
        self.action     = action
        if self.action == 1:
            color_ = 'cyan'
        else:
            color_ = 'yellow'

        self.value_shift = value_shift
        self.annotations = []
        self.x  = None
        self.y  = None
        for ax in ax_list:
            annotation = ax.annotate(self.text_template, 
                        xy=(self.x, self.y), xytext=(self.xoffset, self.yoffset), 
                        textcoords='offset points', ha='right', va='bottom',
                        bbox=dict(boxstyle='round,pad=0.5', fc=color_, alpha=0.5),
                        arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0')
                        )
            annotation.set_visible(False)
            self.annotations.append( annotation )

    def __call__( self, event_type, event, im_array_list=[], im_shift_list=[] ):

        if event_type == 1:
            if event.key == 'escape':
                for annotation in self.annotations:
                    annotation.set_visible(False)
                event.canvas.draw()
                return

            if self.x == None:
                return

            if event.key == 'up':
                self.y -= 1
            elif event.key == 'down':
                self.y += 1
            elif event.key == 'right':
                self.x += 1
            elif event.key == 'left':
                self.x -= 1
            else:
                # すなわち、矢印キー以外
                return

        else:
            # event_type == 3

            if not get_shift_key_state():
                return

            # すなわち、shift + button_press のとき、
            # 新しい位置に設定する。
            self.x, self.y = event.xdata, event.ydata

        # 以下は、新しい位置に設定するか、矢印キーで移動した場合の（再）描画

        y_limit, x_limit = im_array_list[0].shape

        # 注記の表示位置を調整するための関数
        def compute_offsets( ax, x, y ):
            xmin, xmax = ax.get_xlim()
            if (x - xmin)*x_limit/(xmax - xmin) < 200:
                xoffset = 50
            else:
                xoffset = -20
            return ( xoffset, 20 )

        i = 0
        for annotation in self.annotations:
            # print( 'get_xy_limits=', ( self.ax_list[i].get_xlim(), self.ax_list[i].get_ylim() ) )
            if self.action == 1:
                shift_ = im_shift_list[i]
                if len( shift_ ) == 2:
                    x_, y_ = int(self.x + shift_[1]), int(self.y + shift_[0])
                else:
                    x_, y_ = int(self.x), int(self.y)
            else:
                x_, y_ = self.x, self.y
            if x_ < x_limit and y_ < y_limit:
                intensity = im_array_list[i][y_, x_] - self.value_shift
                annotation.xy = x_, y_
                annotation.set_text(self.text_template % ( x_, y_, intensity ) )
                annotation.set_visible(True)
                annotation.set_position( compute_offsets( self.ax_list[i], x_, y_ ) )
            else:
                annotation.set_visible(False)
            i += 1
        event.canvas.draw()

class ColorBar:
    def __init__( self, im, ax ):

        ticks = np.logspace( 1.0, 6.0, num=6 )
        string_labels = []
        for t in range(1,7):
            string_labels.append( r"$10^{%d}$" % t )

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="4%", pad=0.05 )

        self.cb = plt.colorbar( im, cax, ticks=ticks )
        self.cb.set_ticklabels( string_labels )

class MplBackGround:
    def __init__(self):
        plt.style.use('dark_background')

    def __del__(self):
        pass
        # plt.style.use('default')
        # print('MplBackGround.__del__')
        # instead, explicitly use reset_to_default_style below

def reset_to_default_style():
    # print('reset_to_default_style')
    plt.style.use('default')
    mpl_1_5_backward_compatible_init()



# coding: utf-8
"""

    TkMiniTable.py

    Copyright (c) 2016-2020, Masatsuyo Takahashi, KEK-PF

"""
from __future__ import division, print_function, unicode_literals
import sys
import numpy                as np
if sys.version_info > (3,):
    import tkinter as Tk
    import tkinter.ttk as ttk
else:
    import Tkinter as Tk
    import ttk

try:
    import pyperclip
    pyperclip_installed = True
except:
    pyperclip_installed = False

DEBUG = False

class Cell:
    def __init__( self, **entries ): 
        self.__dict__.update( entries )

    def __str__( self ):
        return 'TkMiniTable.Cell<' + str( (self.row, self.col) ) + '>'

class TkMiniTable( Tk.Frame ):
    def __init__( self, parent, columns=[],
        label_bg='gray40', label_fg='white',
        selected_bg='steelblue', selected_fg='white',
        default_colwidth=8,
        font=None,
        *args, **kw ):

        self.cell_widget_class = kw.pop('cell_widget_class', Tk.Label)
        self.on_cell_click = kw.pop('on_cell_click', None)
        self.read_only = kw.pop('read_only', False)
        self.on_cell_click_only = kw.pop('on_cell_click_only', False)

        Tk.Frame.__init__( self, parent, *args, **kw )

        self.parent         = parent
        self.label_bg       = label_bg
        self.label_fg       = label_fg
        self.selected_bg    = selected_bg
        self.selected_fg    = selected_fg
        self.default_colwidth   = default_colwidth
        self.font           = font
        self.active_cell    = None

        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=1)
        # self.columnconfigure(0, weight=1)

        self.L_frame = Tk.Frame( self )
        self.R_frame = Tk.Frame( self )
        self.L_frame.grid( row=0, column=0, sticky="nsew" )
        self.R_frame.grid( row=0, column=1, sticky="nsew" )

        """
            L_frame      R_frame
              L_frame1
              L_frame2
        """

        # self.L_frame.rowconfigure(0, weight=1)
        self.L_frame.columnconfigure(0, weight=1)
        self.L_frame.rowconfigure(1, weight=1)

        self.R_frame.rowconfigure(0, weight=1)

        self.L_frame1 = Tk.Frame( self.L_frame )
        self.L_frame1.grid( row=0, column=0, sticky="nsew" )

        vscrollbar = Tk.Scrollbar( self.R_frame, orient=Tk.VERTICAL )
        vscrollbar.grid( row=0, column=0, sticky="ns" )

        self.canvas = canvas = Tk.Canvas( self.L_frame, bd=0,
                                        highlightthickness=0,
                                        yscrollcommand=vscrollbar.set,
                                        )
        canvas.grid( row=1, column=0, sticky="nsew" )
        vscrollbar.config( command=canvas.yview )
        canvas.yview_moveto(0)

        # create a frame inside the canvas which will be scrolled with it
        self.interior = interior = Tk.Frame( canvas )
        interior_id = canvas.create_window(0, 0, window=interior,
                                           anchor=Tk.NW)

        # track changes to the canvas and frame width and sync them,
        # also updating the scrollbar
        def _configure_interior( canvas, interior, event ):
            # update the scrollbars to match the size of the inner frame
            canvas.config( scrollregion=( 0, 0, interior.winfo_reqwidth(), interior.winfo_reqheight() ) )
            if interior.winfo_reqwidth() != canvas.winfo_width():
                # update the canvas's width to fit the inner frame
                canvas.config( width=interior.winfo_reqwidth() )
            if interior.winfo_reqheight() != canvas.winfo_height():
                # update the canvas's height to fit the inner frame
                canvas.config( height=interior.winfo_reqheight() )

        interior.bind('<Configure>', lambda event: _configure_interior( canvas, interior, event ) )

        # self.iframe = Tk.Frame( self )
        # self.iframe.pack( fill=Tk.BOTH, expand=1 )
        self.L_frame2   = interior
        self.columns    = columns

        self.cells_array = None

        self.reset_cells()
        self.bind( "<Control-c>", self._on_control_c )  # doesn't work

    def add_mousewheel_bind( self, top_widget ):
        self.top_widget = top_widget
        self.bind_all("<MouseWheel>", self._on_mousewheel)

    def reset_cells( self ):

        if self.cells_array is not None:
            for i in range( self.cells_array.shape[0] ):
                for j in range( self.cells_array.shape[1] ):
                    self.cells_array[i, j].widget.destroy()

        self.widgets_dict = {}
        self.selected_cells = []
        self.selected_rows  = []

        _cells = []
        self.columns_dict  = {}
        self.columns_array = []
        for j, col in enumerate( self.columns ):
            svar = Tk.StringVar()
            svar.set( col )

            w = self.cell_widget_class( self.L_frame1, textvariable=svar, relief=Tk.RIDGE, borderwidth=1,
                            width=self.default_colwidth,
                            font=self.font,
                            bg=self.label_bg, fg=self.label_fg )
            w.grid( row=0, column=j )
            w.bind( '<Button-1>', self._on_click )
            # w.bind( '<Control-c>', self._on_control_c )

            cell = Cell( row=0, col=j, svar=svar, widget=w, justify=Tk.CENTER, width=self.default_colwidth,
                        font=self.font,
                        inactive_bg=self.label_bg, inactive_fg=self.label_fg
                        )

            _cells.append( cell )

            head = Cell( col=j, justify=Tk.CENTER, width=self.default_colwidth, font=self.font,
                        bg='whitesmoke', fg='black'
                        )

            self.columns_dict[ col ] = head
            self.columns_array.append( head )

            self.widgets_dict[ w ] = cell

        self.cells_array    = np.array( [ _cells ] )

    def heading( self, colid, text=None, justify=Tk.CENTER ):
        j = self.columns_dict[ colid ].col
        cell = self.cells_array[ 0, j ]
        cell.svar.set( text )
        cell.widget.config( justify=justify )

    def column( self, colid, justify=Tk.CENTER, width=None, bg=None, **kwargs ):
        try:
            # if colid is iterable loop as expected.
            for i in colid:
                self.column_impl( i, justify, width, bg, **kwargs )
            return
        except Exception as exc:
            if DEBUG:
                raise exc
            pass

        self.column_impl(colid, justify, width, bg, **kwargs)

    def column_impl( self, colid, justify, width, bg, **kwargs):

        if type( colid ) == int:
            column = self.columns_array[ colid ]
        else:
            column = self.columns_dict[ colid ]

        if width is None:
            width = self.default_colwidth
        # print( colid, 'column width=', width )
        column.width = width
        if bg is not None:
            column.bg = bg

        n_ = column.col
        # print( 'column: n_=', n_ )
        column_cells = self.cells_array[ :, n_ ]
        for cell in column_cells:
            kwargs = { 'justify':justify, 'width':width }
            cell.widget.config( **kwargs )
            cell.width = width
        self.update()

    def clear_rows( self ):
        pass

    def insert_row( self, row_tuple ):
        id_ = self.cells_array.shape[0]

        _cells = []
        for j, col in enumerate( row_tuple ):

            column = self.columns_array[j]

            svar = Tk.StringVar()
            svar.set( col )

            if j == 0:
                bg = self.label_bg
                fg = self.label_fg
            else:
                # bg = 'whitesmoke'
                # fg = 'black'
                bg = column.bg
                fg = column.fg

            w = self.cell_widget_class( self.L_frame2, textvariable=svar,
                            justify=column.justify, width=column.width, relief=Tk.RIDGE, borderwidth=1,
                            font=self.font,
                            bg=bg, fg=fg )

            w.grid( row=id_-1, column=j )
            w.bind( '<Button-1>', self._on_click )
            w.bind( '<Button-3>', self._on_right_click )
            w.bind( '<Double-Button-1>', self._on_double_click )
            # w.bind( '<Control-c>', self._on_control_c )

            cell = Cell( row=id_, col=j, svar=svar, widget=w, inactive_bg=bg, inactive_fg=fg )
            _cells.append( cell )

            self.widgets_dict[ w ] = cell

        self.cells_array = np.vstack( [ self.cells_array, np.array( [ _cells ] ) ] )

        # print( 'cells_array.shape=', self.cells_array.shape )

        return _cells

    def _select_row( self, row ):
        if len( self.selected_rows ) > 0:
            prev_row = self.selected_rows[0]
            if row == prev_row: return

        for cell in self.selected_cells:
            cell.widget.config( bg=cell.inactive_bg, fg=cell.inactive_fg )

        if row == 0:
            self.selected_rows  = []
            return

        self.selected_rows  = [ row ]

        self.selected_cells = []
        row_cells = self.cells_array[ row, : ]
        for cell in row_cells:
            bg_ = cell.widget.cget( 'bg' )
            cell.inactive_bg = bg_
            fg_ = cell.widget.cget( 'fg' )
            cell.inactive_fg = fg_
            cell.widget.config( bg=self.selected_bg, fg=self.selected_fg )
            self.selected_cells.append( cell )

    def activate_cell( self, cell ):
        if DEBUG: print( 'activate_cell', cell )
        if self.active_cell is not None:
            a_cell = self.active_cell
            if a_cell.row in self.selected_rows:
                bg  = self.selected_bg
                fg  = self.selected_fg
            else:
                bg  = a_cell.inactive_bg
                fg  = a_cell.inactive_fg
            self.active_cell.widget.config( bg=bg, fg=fg )

        self.active_cell = cell
        if DEBUG: print( 'activate_cell active_cell=', self.active_cell )
        cell.widget.config( bg='blue', fg='white' )
        cell.widget.focus_force()

    def _on_click( self, event ):
        if DEBUG: print( '_on_click event.widget=', event.widget )
        if self.read_only:
            return

        cell = self.widgets_dict.get( event.widget )
        if cell is None:
            if DEBUG: print( '_on_click active_cell=', self.active_cell )
            self.inactivate()
            return

        if self.on_cell_click is not None:
            self.on_cell_click(cell)
            if self.on_cell_click_only:
                return

        # print( 'click', [ cell.row, cell.col ] )
        self._select_row( cell.row )

        self.activate_cell( cell )

        if hasattr( self.parent, '_on_click' ):
            self.parent._on_click( event )

    def select_row( self, row ):
        row_ = row + 1
        self._select_row(row_)
        cell = self.cells_array[ row_, 1 ]
        self.activate_cell(cell)
        move_y = max(0, row_ - 5) / self.cells_array.shape[0]
        self.canvas.yview_moveto(move_y)

    def inactivate( self ):
        if DEBUG: print( 'inactivate active_cell=', self.active_cell )

        if self.active_cell is None: return

        a_cell = self.active_cell
        if a_cell.row == 0 or a_cell.col == 0:
            bg  = self.label_bg
            fg  = self.label_fg
        else:
            bg  = self.selected_bg
            fg  = self.selected_fg
        self.active_cell.widget.config( bg=bg, fg=fg )
        self.active_cell = None

    def _on_double_click( self, event ):
        cell = self.widgets_dict.get( event.widget )
        if cell is None: return

        # print( 'double click', [ cell.row, cell.col ] )

        if hasattr( self.parent, '_on_double_click' ):
            self.parent._on_double_click( event )

    def _on_right_click( self, event ):
        if len( self.selected_rows ) == 0: return

        if hasattr( self.parent, '_on_right_click' ):
            self.parent._on_right_click( event )

    def _on_control_c( self, event ):
        # print( '_on_control_c' )
        if not pyperclip_installed:
            print( 'Install pyperclip if you need to copy to the clipboard.' )
            return
        if self.active_cell is None: return

        value = self.active_cell.svar.get()
        pyperclip.copy( value )
        print( 'active cell value copied to the clipboard', value )

    def _on_mousewheel( self, event ):
        """
        scroll only if the window is in focus
        TODO: any other way?
        
        http://stackoverflow.com/questions/10343759/determining-what-tkinter-window-is-currently-on-top
        Determining what tkinter window is currently on top
        """
        num_stacked = len( self.tk.eval( 'wm stackorder '+str(self.top_widget) ).split(' ') )
        if num_stacked == 1:
            self.canvas.yview_scroll( int( -1*(event.delta/120) ), "units")



"""
    FoleyDorseyEmg.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, root
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Peaks.ElutionModels import egh, emg
from LPM import get_corrected
from .BoundControl import Penalties
from .ColumnConstants import Ti_LOWER

ESTIMATE_INJECTION_TIME = True

def _pre_optimize_params(ecurve, xr_params, seccol_params, debug=False):
    x = ecurve.x
    y_ = ecurve.y

    # use corrected y
    y = get_corrected(y_)

    # Npc, rp, tI, t0, P, m
    Npc, tI = seccol_params[[0,2]]

    n = xr_params.shape[0]
    temp_xr_params = np.zeros((n,3))

    for k, (h, m, t) in enumerate(xr_params[:,[0,1,3]]):
        temp_xr_params[k,0] = h
        temp_xr_params[k,1] = m
        temp_xr_params[k,2] = t

    def objective(p):
        N_, T_ = p
        sqN = np.sqrt(N_)
        my = np.zeros(len(x))
        for k, (h, m, t) in enumerate(temp_xr_params):
            s = compute_sigma(T_, N_, m, t)
            my += emg(x, h, m, s, t)
        return np.sum((my - y)**2)

    ret = minimize(objective, (Npc, tI))

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        print("tI=", tI)

        with plt.Dp():
            fig, (ax0, ax1, ax2) = plt.subplots(ncols=3, figsize=(18,5))
            fig.suptitle("Foley-Dorsey Adaption Proof", fontsize=20)
            ax0.set_title("Free EGH Decomposition", fontsize=16)
            ax1.set_title("Foley-Dorsey Simple Conversion", fontsize=16)
            ax2.set_title("Foley-Dorsey Optimized Adaption", fontsize=16)

            ax0.plot(x, y)
            for k, (h, m, s, t) in enumerate(xr_params):
                ax0.plot(x, egh(x, h, m, s, t), ":")

            def plot_params(ax, tI, Npc):
                ax.plot(x, y)
                for k, (h, m, t) in enumerate(temp_xr_params):
                    s = compute_sigma(tI, Npc, m, t)
                    print([k], h, m, s, t)
                    ax.plot(x, emg(x, h, m, s, t), ":")
            plot_params(ax1, tI, Npc)
            plot_params(ax2, *ret.x)

            fig.tight_layout()
            plt.show()

    return temp_xr_params, ret.x

def fd_estimate_initial_params(ecurve, init_xr_params, seccol_params, debug=False):
    temp_xr_params, (N_, T_) = _pre_optimize_params(ecurve, init_xr_params, seccol_params, debug=debug)
    fd_seccol_params = seccol_params.copy()
    fd_seccol_params[[0,2]] = N_, T_
    return temp_xr_params, fd_seccol_params

def compute_tau_simple(tI, Npc, mu, sigma):
    t = (mu - tI)**2/Npc - sigma**2
    return np.sqrt(t) if t > 0 else 0

def compute_tau(tI, Npc, mu, sigma):
    """
    from sympy.abc import m, s, t, N
    solve([(m+t)**2 - N*(s**2+t**2)], [t])
    [(m/(N - 1) - sqrt(-N*(N*s**2 - m**2 - s**2))/(1 - N),),
     (m/(N - 1) + sqrt(-N*(N*s**2 - m**2 - s**2))/(1 - N),)]
    """
    m = mu - tI
    t2 = Npc*(m**2 + (1 - Npc)*sigma**2)

    if np.isscalar(t2):
        if t2 >= 0:
            t_ = np.sqrt(t2)
        else:
            if fronting_ok:
                t_ = np.sqrt(-t2)
            else:
                t_ = 0
                Penalties[0] += abs(t2)
    else:
        # vectors will be used in ModelParams/FdEmgParams.py
        if np.all(t2 >= 0):
            t_ = np.sqrt(t2)
        else:
            t_ = np.zeros(len(t2))
            Penalties[0] = np.sum(np.abs(np.min([t_, t2], axis=0)))
            t_[t2 > 0] = np.sqrt(t2[t2 > 0])
    return (m + t_)/(Npc - 1)

def compute_sigma(tI, Npc, mu, tau):
    return np.sqrt((mu + tau - tI)**2/Npc - tau**2)

def convert_to_xr_params_fd(c_xr_params, tI, Npc):
    nc = c_xr_params.shape[0]
    xr_params = np.zeros((nc, 4))
    sqrtNp = np.sqrt(Npc)
    Penalties[0] = 0
    for k, (h, m, t) in enumerate(c_xr_params):
        xr_params[k,0] = h
        xr_params[k,1] = m
        xr_params[k,2] = compute_sigma(tI, Npc, m, t)
        xr_params[k,3] = t
    return xr_params

def spike_demo():
    tI = -2000

    x = np.linspace(0, 500, 501)

    t0 = 50
    mu = 200
    sigma = 30
    tau = 40
    y = emg(x, 1, mu, sigma, tau)

    tr = mu + tau
    N = (tr - tI)**2/(sigma**2 + tau**2)
    print("tI=", tI)
    print("N=", N)

    tau_ = compute_tau(tI, N, mu, sigma)
    print("tau_=", tau_)    # tau == 10

    tau_ = compute_tau(np.array([tI, tI]), np.array([N, N]), np.array([mu, mu]), np.array([sigma, sigma]))
    print("tau_=", tau_)    # tau == 10

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.plot(x, y)
        ax.plot(tr, emg(np.array([tr]), 1, mu, sigma, tau), "o", color="yellow")
        ax.plot(mu, emg(np.array([mu]), 1, mu, sigma, tau), "o", color="pink")
        fig.tight_layout()
        plt.show()

if __name__ == '__main__':
    spike_demo()



# coding: utf-8
"""
    Rgg.RggEghDistribution.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
import scipy.stats as stats
import seaborn as sns
from pomegranate import GaussianKernelDensity
from .RggUtils import normal_pdf
from .KernelDensityFix import UniformKernelDensity
from molass_legacy.Peaks.ElutionModels import egh_pdf, compute_egh_params

import molass_legacy.KekLib.DebugPlot as plt

USE_KERNEL_DENSITY = True
TRIM_IMPOSSILE = False
KD_BANDWIDTH_RATIO = 1
# KD_DISTRIBUTION = GaussianKernelDensity
KD_DISTRIBUTION = UniformKernelDensity
VERY_SMALL_VALUE = 1e-10
USE_SIMPLE_PRODUCT = False

class RggEghDistribution:
    def __init__(self, tR, sigma, tau=0, rg=30, use_copula=True):
        print("use_copula=", use_copula)
        self.name = 'RggEghDistribution'
        self.frozen = False
        self.tR = tR
        self.sigma = sigma
        self.tau = tau
        self.rg = rg
        self.rg_sigma = KD_BANDWIDTH_RATIO
        self.parameters = (self.tR, self.sigma, self.tau, self.rg, self.rg_sigma)
        if USE_KERNEL_DENSITY:
            self.rg_dist = KD_DISTRIBUTION([self.rg], bandwidth=self.rg_sigma)

        self.d = 2
        self.summaries = np.zeros(6)
        if use_copula:
            self.norm_ppf = stats.distributions.norm().ppf
            self.copula = stats.multivariate_normal(mean=[0, 0], cov=[[1., 0.1], 
                                                                      [0.1, 1.]])
        self.use_copula = use_copula
        self.debug_weights = False
        self.debug_probability = False
        self.update_area_params()

    def to_dict(self):
        return {
            'class' : 'Distribution',
            'name'  : self.name,
            'parameters' : self.parameters,
            'frozen' : self.frozen
        }

    def update_area_params(self):
        # self.egh_limits = (self.tR - self.sigma*3, self.tR + self.sigma*3)
        self.rg_limits = (self.rg - self.rg_sigma, self.rg + self.rg_sigma)
        self.ehg_scale = 1/(self.rg_sigma*2)

    def probability(self, X):
        if self.use_copula:
            P0 = egh_pdf(X[:,0], self.tR, self.sigma, self.tau)
            if USE_KERNEL_DENSITY:
                X1 = X[:,1]
                P1 = self.rg_dist.probability(X1)
                if USE_SIMPLE_PRODUCT:
                    return P0*P1

                if TRIM_IMPOSSILE:
                    invalid = np.logical_or(X1 < self.rg_limits[0], X1 > self.rg_limits[1])
                    P0[invalid] = 0
            else:
                P1 = normal_pdf(X[:,1], self.rg, self.rg_sigma)
            N0 = self.norm_ppf(P0)
            N1 = np.zeros(X.shape[0])
            positive = P1 > VERY_SMALL_VALUE
            N1[positive] = self.norm_ppf(P1[positive])
            if False:
                plt.push()
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,7))
                ax1.plot(X[:,0])
                ax2.plot(X[:,1])
                plt.show()
                plt.pop()
                plt.push()
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,7))
                ax1.plot(P0, P1)
                ax2.plot(N0, N1)
                plt.show()
                plt.pop()
            P_ = self.copula.pdf(np.array([N0, N1]).T)
        else:
            P0 = egh_pdf(X[:,0], self.tR, self.sigma, self.tau)
            X1 = X[:,1]
            P1 = self.rg_dist.probability(X1)
            invalid = np.logical_or(X1 < self.rg_limits[0], X1 > self.rg_limits[1])
            P0[invalid] = 0
            P_ = self.ehg_scale * P0

        if self.debug_probability:
            plt.push()
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            bb_ = np.zeros(X.shape[0])
            ax.bar3d(X[:,0], X[:,1], bb_, 1, 0.1, P_, shade=True, edgecolor='green')
            ax.set_ylim(0, 50)
            fig.tight_layout()
            self.debug_probability = plt.show()
            plt.pop()
        return P_

    def log_probability(self, X):
        p = self.probability(X)
        p[p < VERY_SMALL_VALUE] = VERY_SMALL_VALUE
        return np.log(p)

    def summarize(self, X, w=None):
        if w is None:
            w = numpy.ones(X.shape[0])

        if self.debug_weights:
            plt.push()
            fig, ax = plt.subplots()
            ax.plot(w)
            ret = plt.show()
            plt.pop()
            self.debug_weights = ret

        X_ = X[:,0]
        self.summaries[0] += w.sum()
        if np.isnan(self.summaries[0]):
            raise StopIteration("summaries have included NaN's")

        w_ = w/self.summaries[0]
        self.summaries[1] += X_.dot(w_)
        self.summaries[2] += ((X_ - self.summaries[1])** 2).dot(w_)
        self.summaries[3] += ((X_ - self.summaries[1])** 3).dot(w_)
        X1 = X[:,1]
        self.summaries[4] += X1.dot(w_)
        self.summaries[5] += ((X1 - self.summaries[4])**2).dot(w_)
        print("self.summaries=", self.summaries)

    def from_summaries(self, inertia=0.0):
        init_params = (self.summaries[1], np.sqrt(self.summaries[2]), 0)
        self.tR, self.sigma, tau = compute_egh_params(init_params, self.summaries[1:4])
        moderate_tau = min(self.sigma, max(-self.sigma, tau))
        self.tau = moderate_tau
        self.rg = self.summaries[4]
        self.rg_sigma = np.sqrt(self.summaries[5])
        if USE_KERNEL_DENSITY:
            self.rg_dist = KD_DISTRIBUTION([self.rg], bandwidth=self.rg_sigma*KD_BANDWIDTH_RATIO)
        self.parameters = (self.tR, self.sigma, moderate_tau, self.rg, self.rg_sigma)
        print("self.parameters=", self.parameters)
        self.update_area_params()
        self.clear_summaries()

    def clear_summaries(self, inertia=0.0):
        self.summaries = np.zeros(6)

    @classmethod
    def from_samples(cls, X, weights=None):
        d = RggEghDistribution(100, 30, 0, 23)
        d.summarize(X, weights)
        d.from_summaries()
        return d

    @classmethod
    def blank(cls):
        return RggEghDistribution(100, 30, 0, 23)

def demo():
    x = np.linspace(0, 400, 100)
    dx = (x[1] - x[0])/4
    y = np.linspace(0, 50, 100)
    dy = (y[1] - y[0])/4
    xx, yy = np.meshgrid(x, y)

    xx_ = xx.flatten()
    yy_ = yy.flatten()
    X = np.array([xx_, yy_]).T
    bb_ = np.zeros(len(xx_))

    fig = plt.figure(figsize=(21,7))
    ax0 = fig.add_subplot(131)
    ax1 = fig.add_subplot(132, projection='3d')
    ax2 = fig.add_subplot(133, projection='3d')
    d1 = RggEghDistribution(200, 30, tau=10, rg=30, use_copula=True)
    d2 = RggEghDistribution(200, 30, tau=10, rg=30, use_copula=False)

    ax0.plot(x, egh_pdf(x, 200, 30, 10))

    for ax, d in [(ax1, d1), (ax2, d2)]:
        dz = d.probability(X)
        print("dz.shape=", dz.shape)
        ax.set_xlim(x[0], x[-1])
        ax.bar3d(xx_, yy_, bb_, dx, dy, dz, shade=True, edgecolor='green')

    fig.tight_layout()
    plt.show()



"""
    SvdDenoise.py

    Copyright (c) 2018-2020, 2025, SAXS Team, KEK-PF
"""
import numpy as np
from molass.LowRank.LowRankInfo import get_denoised_data     # for forward compatibility

def get_denoised_error(M_, M, E, left=True, debug=False, q=None, eno=None ):

    """
        M_ = T @ M
        T = M_ @ Mpinv
    """

    Mpinv = np.linalg.pinv( M )
    if left:
        W = np.dot( M_, Mpinv )
        E_ = np.sqrt( np.dot( W**2, E**2) )
    else:
        W = np.dot( Mpinv, M )
        E_ = np.sqrt( np.dot( E**2, W**2) )

    if debug and eno is not None:
        debug_plot(q, M, E, M_, E_, eno=eno)

    return E_

def debug_plot(x, M, E, M_, E_, eno):
    import molass_legacy.KekLib.DebugPlot as plt
    fig = plt.figure(figsize=(21,7))
    ax0  = fig.add_subplot(131)
    ax1  = fig.add_subplot(132)
    ax2  = fig.add_subplot(133)

    ax0.set_title("Error propagation in denoising with SVD using MR-inverve")
    ax0.axes.get_xaxis().set_ticks([])
    ax0.axes.get_yaxis().set_ticks([])
    ax0.text(0.1,0.85, r"$M = U \cdot \Sigma \cdot V^T$", fontsize=30)
    ax0.text(0.1,0.65, r"$\widetilde{M} = \widetilde{U} \cdot \widetilde{\Sigma} \cdot  \widetilde{V}^T $", fontsize=30)
    ax0.text(0.1,0.45, r"$\widetilde{M} \approx  W \cdot M$", fontsize=30)
    ax0.text(0.1,0.25, r"$W \approx  W \cdot M \cdot M^+ \approx  \widetilde{M} \cdot M^+$", fontsize=30)
    ax0.text(0.1,0.05, r"$\widetilde{E} = \sqrt{ W^{(2)} \cdot E^{(2)} }$", fontsize=30)
    """
        [ a b ]  [x]
        [ c d ]  [y]
        ax + by
        cx + dy
    """

    ax1.set_title( "Propagated errors in linear scale at eno=%d" % eno )
    ax1.set_xlabel('Q')
    ax1.set_ylabel('Intensity')
    ax1.plot(x, M[eno, :], label='data')
    ax1.plot(x, E[eno, :], label='error')
    ax1.plot(x, M_[eno, :], label='denoised data')
    ax1.plot(x, E_[eno, :], label='denoised error')
    ax1.legend()
    ax2.set_title( "Propagated errors in $Log_{10}$ scale at eno=%d" % eno )
    ax2.set_xlabel('Q')
    ax2.set_ylabel('$Log_{10}(Intensity)$')
    ax2.plot(x, np.log10(M[eno, :]), label='data')
    ax2.plot(x, np.log10(E[eno, :]), label='error')
    ax2.plot(x, np.log10(M_[eno, :]), label='denoised data')
    ax2.plot(x, np.log10(E_[eno, :]), label='error of denoised data')
    ax2.legend()
    fig.tight_layout()
    plt.show()



# -*- coding: utf-8 -*-
"""

    ファイル名：   ControlKeyState.py

    処理内容：

       shift キーと ctrl キーの状態管理

"""
from __future__ import division, print_function, unicode_literals

ctrl_is_held = None

def set_ctrl_key_state( state ):
    global ctrl_is_held
    ctrl_is_held = state

def get_ctrl_key_state():
    global ctrl_is_held
    return ctrl_is_held

shift_is_held = None

def set_shift_key_state( state ):
    global shift_is_held
    shift_is_held = state

def get_shift_key_state():
    global shift_is_held
    return shift_is_held



"""

    ReportUtils.py

    Copyright (c) 2016-2024, SAXS Team, KEK-PF

"""
import os
import numpy as np
from molass_legacy.AutorgKek.Quality import fit_consistency_func, compute_atsas_fit_consistency
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.KekLib.NumpyUtils import np_savetxt, np_loadtxt
from molass_legacy.KekLib.ProgressInfo import put_info
from molass_legacy.SerialAnalyzer.ProgressInfoUtil import STREAM_GUINIER
from .GuinierAnalysisResultBook import GuinierAnalysisResultBook
from .ZeroExtrapolationResultBook import ZeroExtrapolationResultBook
from .ZeroExtrapolationOverlayBook import ZeroExtrapolationOverlayBook
from .SummaryBook import SummaryBook
from molass.Reports.Migrating import COLNAMES, make_gunier_row_values

BASIC_QUALITY   = 2
Q_RG_SCORE      = 5
I0          = 10
RG          = 11
GPFIT_RG    = 14
GPFIT_I0    = 19
I0_STDEV    = 27
RG_STDEV    = 28
ATSAS_BEGIN = 29
ATSAS_I0    = 29
ATSAS_RG    = 30
ATSAS_QUALITY   = 32

SELECT_COLUMS = [ 2, 3, 5, 6, 7, ATSAS_QUALITY, 10, GPFIT_I0, ATSAS_I0, 11, GPFIT_RG, ATSAS_RG ]

def toFloat( value ):
    return 0 if value == '' else float( value )

class RestoredResult:
    def __init__( self, saved_line ):
        rec_row = saved_line.split( ',' )

        self.I0         = toFloat( rec_row[I0] )
        self.I0_        = self.I0
        self.I0_stdev   = toFloat( rec_row[I0_STDEV] )
        self.Rg         = toFloat( rec_row[RG] )
        self.Rg_stdev   = toFloat( rec_row[RG_STDEV] )
        self.Rg_        = self.Rg
        self.gpfit_Rg   = toFloat( rec_row[GPFIT_RG] )
        self.gpfit_I0   = toFloat( rec_row[GPFIT_I0] )

        # TODO: add From, To
        self.From       = 0     # TBD
        self.To         = 0     # TBD

def get_header_record():
    return ','.join( COLNAMES )

def make_record( path, result, result_atsas, result_atsas_eval=None ):
    results = make_gunier_row_values(result, result_atsas)
    # print( 'Rg=', result.Rg )

    def convert_to_float( x ):
        if x is None:
            x = 0

        try:
            xstr = '%g' % x
        except:
            xstr = 'NA'

        return xstr

    folder, file = os.path.split( path )

    rec = ','.join( [ folder, file ]
                    + [ '%g' % ( 0 if x is None else x) for x in results ] )
                    # + list( map( convert_to_float, results ) ) )

    return rec

def load_guinier_result( result_file, mc_vector ):
    fh = open( result_file )
    array = []
    for j, line in enumerate(fh, start=-1):
        if j == -1:
            # skip header record
            continue
        in_row = line.split( ',' )
        if len(in_row) <= ATSAS_BEGIN:
            in_row += [ '' ] * (len(COLNAMES) - len(in_row))
        c = mc_vector[j]
        out_row = in_row[0:2] + [ c ]
        out_row += [ toFloat(in_row[i]) for i in SELECT_COLUMS]

        array.append( out_row )

    fh.close()
    return array

def just_load_guinier_result( result_file ):
    fh = open( result_file )
    array = []
    for j, line in enumerate(fh, start=-1):
        if j == -1:
            # skip header record
            continue
        in_row = line.split( ',' )
        if len(in_row) <= ATSAS_BEGIN:
            in_row += [ '' ] * (len(COLNAMES) - len(in_row))

        out_row = in_row[0:2]
        out_row += [ toFloat(v) for v in in_row[2:] ]

        array.append( out_row )

    fh.close()
    return array


def make_guinier_analysis_book(
            excel_is_available, wb, ws,
            book_file, serial_file, mc_vector, j0,
            applied_ranges=None,
            cmd_queue=None, parent=None, logger=None ):

    assert( applied_ranges is not None )

    ranges = []
    for range_ in applied_ranges:
        fromto_list = range_.get_fromto_list()
        ranges.append( [ fromto_list[0][0], fromto_list[-1][1] ] )

    put_info( (STREAM_GUINIER, 1), 1 )

    array = load_guinier_result( serial_file, mc_vector )
    put_info( (STREAM_GUINIER, 1), 2 )

    book = GuinierAnalysisResultBook( wb, ws, array, j0, parent=parent )
    put_info( (STREAM_GUINIER, 1), 3 )

    if excel_is_available:
        book.save( book_file )

    put_info( (STREAM_GUINIER, 1), 4 )

    if excel_is_available:
        book.add_annonations( book_file, ranges )
    else:
        if logger is not None:
            _, file = os.path.split(book_file)
            logger.warning( 'excel book formatting has been skipped for %s.', file )

    put_info( (STREAM_GUINIER, 1), 5 )

def make_zero_extrapolation_book(
            excel_is_available, wb, ws,
            book_file, mjn, indeces, c_rg_iz, qvector, data_matrix, j0,
            ze_array, param_array, orig_param_array,
            guinier_boundary,
            almerge_result, atsas_result,
            boundary_j, parent=None, lrf_info=None,
            logger=None ):

    prefix = book_file.replace( '.xlsx', '' )
    # np_savetxt( prefix + '-data_matrix.csv', data_matrix )
    # np_savetxt( prefix + '-ze_array.csv', ze_array )
    book = ZeroExtrapolationResultBook( wb, ws, mjn, indeces, c_rg_iz, qvector, data_matrix, j0,
                                        ze_array, param_array, orig_param_array,
                                        guinier_boundary,
                                        almerge_result, atsas_result,
                                        boundary_j, parent=parent, lrf_info=lrf_info )
    if excel_is_available:
        book.save( book_file )
        book.add_format_setting( book_file )
    else:
        if logger is not None:
            _, file = os.path.split(book_file)
            logger.warning( 'excel book formatting has been skipped for %s.', file )

def dump_zx_report_args( m, j, indeces, c_rg_iz, qvector, data_matrix, ze_array, param_array, orig_param_array, exz_array ):
    np_savetxt( 'indeces-%d-%d.csv' % (m, j), indeces )
    for k, list_k in enumerate( c_rg_iz ):
        for n, array in enumerate( list_k ):
            # print( (k,n), 'array.shape=', array.shape )
            # if (k,n) == (1,1):
            #    print( 'array=', array )
            if len( array.shape ) == 1:
                array_ = [ np.nan if a is None else a for a in array ]
            elif len( array.shape ) == 2:
                array_ = np.array( array )
                for p in range( array.shape[0] ):
                    for q in range( array.shape[1] ):
                        if array[p,q] is None:
                            array_[p,q] = np.nan
            else:
                assert( False )
            np_savetxt( 'c_rg_iz-%d-%d-%d-%d.csv' % (m, j, k, n), np.array(array_) )
    np_savetxt( 'qvector-%d-%d.csv' % (m, j), qvector )
    np_savetxt( 'data_matrix-%d-%d.csv' % (m, j), data_matrix )
    np_savetxt( 'ze_array-%d-%d.csv' % (m, j), ze_array )
    np_savetxt( 'param_array-%d-%d.csv' % (m, j), param_array )
    if orig_param_array is not None:
        np_savetxt( 'orig_param_array-%d-%d.csv' % (m, j), orig_param_array )
    if exz_array is not None:
        np_savetxt( 'exz_array-%d-%d.csv' % (m, j), exz_array )

def load_zx_report_args( folder, m, j ):
    indeces_csv = folder + '/indeces-%d-%d.csv' % ( m, j )
    indeces, _ = np_loadtxt( indeces_csv )

    c_rg_iz = []
    for k in range(3):
        list_ = []
        for n in range(2):
            c_rg_iz_csv =  folder + '/c_rg_iz-%d-%d-%d-%d.csv' % ( m, j, k, n )
            array = np.genfromtxt( c_rg_iz_csv, delimiter=',' )
            if len( array.shape ) == 1:
                array_ = np.array( [ None if np.isnan(a) else a for a in array ] )
            elif len( array.shape ) == 2:
                # array_ = np.array( array )
                array_= array
                for p in range( array.shape[0] ):
                    for q in range( array.shape[1] ):
                        if np.isnan( array[p,q] ):
                            array_[p,q] = None
            else:
                assert( False )
            # print( (k, n), 'array_=', array_ )
            list_.append( array_ )
        c_rg_iz.append( list_ )

    qvector_csv = folder + '/qvector-%d-%d.csv' % ( m, j )
    qvector, _ = np_loadtxt( qvector_csv )

    data_matrix_csv = folder + '/data_matrix-%d-%d.csv' % ( m, j )
    if os.path.exists( data_matrix_csv ):
        data_matrix, _ = np_loadtxt( data_matrix_csv )
    else:
        data_matrix = None

    ze_array_csv = folder + '/ze_array-%d-%d.csv' % ( m, j )
    if os.path.exists( ze_array_csv ):
        ze_array, _ = np_loadtxt( ze_array_csv )
    else:
        ze_array = None

    param_array_csv = folder + '/param_array-%d-%d.csv' % ( m, j )
    # param_array = np_loadtxt( param_array_csv )   # results in an error
    param_array = np.loadtxt( param_array_csv, delimiter=',' )
    print( 'param_array.shape=', param_array.shape )

    orig_param_array_csv = folder + '/orig_param_array-%d-%d.csv' % ( m, j )
    if os.path.exists( orig_param_array_csv ):
        orig_param_array, _ = np_loadtxt( orig_param_array_csv )
    else:
        orig_param_array = None

    exz_array_csv = folder + '/exz_array-%d-%d.csv' % ( m, j )
    if os.path.exists( exz_array_csv ):
        exz_array, _ = np_loadtxt( exz_array_csv )
    else:
        exz_array = None

    return indeces, c_rg_iz, qvector, data_matrix, ze_array, param_array, orig_param_array, exz_array

def make_zero_extrapolation_overlay_book(
            excel_is_available, wb, ws,
            book_file, mn, qvector, param_array_list, boundary_indeces,
            parent=None, lrf_info=None, logger=None ):

    overlay = ZeroExtrapolationOverlayBook( wb, ws, mn, qvector, param_array_list, parent=parent, lrf_info=lrf_info )
    if excel_is_available:
        overlay.save( book_file )
        overlay.add_format_setting( book_file, boundary_indeces=boundary_indeces )
    else:
        if logger is not None:
            _, file = os.path.split(book_file)
            logger.warning( 'excel book formatting has been skipped for %s.', file )

def make_summary_book(excel_is_available, wb, book_file, controller, debug=False):
    if debug:
        from OnTheFly.DebugDialog import DebugDialog
        dialog = DebugDialog(debug_info=[book_file, controller])
        dialog.show()

    summary = SummaryBook( wb, controller )

    if excel_is_available:
        summary.save( book_file )
        summary.add_format_setting( book_file )

    path = book_file.replace( '.xlsx', '.csv' )
    print( 'path=', path )
    summary.save_as_csv( path )



"""
    Optimizer.ParamSetType.EghParamsSheet.py

    Copyright (c) 2021-2023, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk
from tksheet import Sheet
from molass_legacy._MOLASS.SerialSettings import get_setting
from .ParamsSheetBase import ParamsSheetBase

# task : unify parameter names with those in EghParams.py

class EghParamsSheet(ParamsSheetBase):
    def __init__(self, parent, params, dsets, optimizer, model_func=None):
        self.composite = optimizer.composite
        self.model_func = model_func
        ParamsSheetBase.__init__(self, parent, params, dsets, optimizer)
        self.build_sheet(params, dsets, optimizer)

    def get_wanted_params(self, params, optimizer):
        return optimizer.split_params_simple(params) + [[], 0, (None, None)]

    def build_sheet(self, params, dsets, optimizer):
        self.n = optimizer.n_components

        body_frame = Tk.Frame(self)
        body_frame.pack()

        dark_cells = []
        xr_params, xr_baseparams, rgs, mapping, uv_params, uv_baseparams, mappable_range, seccol_params, dark_cell_j_list, num_extra_addresses, (Ti, Np) = self.get_wanted_params(params, optimizer)

        constrained_anyway = num_extra_addresses> 0
        num_addresses = len(params) + num_extra_addresses
        self.params_addr = [None] * num_addresses
        self.params_addr_inv = {}

        xr_base = 0
        xr_bp_base = xr_base + np.product(xr_params.shape)
        rg_base = xr_bp_base + len(xr_baseparams)
        mp_base = rg_base + len(rgs)
        uv_base = mp_base + len(mapping)
        uv_bp_base = uv_base + len(uv_params)
        mr_base = uv_bp_base + len(uv_baseparams)
        se_base = mr_base + len(mappable_range)

        if dsets is None:
            num_columns = 9
            extra_headers = []
            xr_proportions = None
            uv_proportions = None
        else:
            if self.model_func is None:
                from molass_legacy.Peaks.ElutionModels import egh
                model_func = egh
            else:
                model_func = self.model_func
            num_columns = 11
            extra_headers = ["", "", "proportion"]

            opt_lrf_info = optimizer.objective_func(params, return_lrf_info=True)
            xr_proportions = opt_lrf_info.get_xr_proportions()
            uv_proportions = opt_lrf_info.get_uv_proportions()

        if seccol_params is None:
            num_extended_rows = 0
        else:
            num_extended_rows = len(seccol_params) + 1

        data_list = [["" for c in range(num_columns)] for r in range(self.n*2 + 2 + 8 + num_extended_rows)]

        row_offset = 0
        for j, name in enumerate(["h", "mu (tR)", "sigma", "tau", "composite", "", "rg"] + extra_headers, start=1):
            data_list[row_offset][j] = name

        row_offset += 1
        n = xr_params.shape[0]
        xr_col_size = xr_params.shape[1]
        if self.composite is None or not self.composite.really_composite:
            composite_ids = None
            show_flags = None
            proportions = xr_proportions
        else:
            composite_ids = self.composite.get_composite_ids()
            show_flags = self.composite.get_composite_show_flags()
            proportions = self.composite.get_composite_proportions(xr_proportions)

        for i in range(n):
            if i == 0:
                v = "xr_params"
            else:
                v = ""
            data_list[row_offset+i][0] = v
            for j in range(xr_col_size):
                data_list[row_offset+i][j+1] = "%g" % xr_params[i,j]
                k = i*4 + j
                self.set_params_addr(xr_base+k, (row_offset+i, j+1))
                if j in dark_cell_j_list:
                    dark_cells.append((row_offset+i, j+1))

            if composite_ids is not None:
                data_list[row_offset+i][xr_col_size+1] = str(composite_ids[i]+1)

            show_value = show_flags is None or show_flags[i]
            if show_value:
                data_list[row_offset+i][xr_col_size+3] = "%g" % rgs[i]
            self.set_params_addr(rg_base+i, (row_offset+i, xr_col_size+3))

            if xr_proportions is not None:
                if i == 0:
                    data_list[row_offset+i][xr_col_size+5] = "xr area"
                if i < len(xr_proportions):
                    if show_value:
                        i_ = i if composite_ids is None else composite_ids[i]
                        data_list[row_offset+i][xr_col_size+6] = "%g" % proportions[i_]

        row_offset += n + 1
        col_names = ["slope", "intercept"]
        if len(xr_baseparams) == 3:
            col_names.append("fouling")

        for j, name in enumerate(col_names):
            data_list[row_offset][1+j] = name

        row_offset += 1
        for i in range(1):
            if i == 0:
                v = "xr_baseline"
            else:
                v = ""
            data_list[row_offset+i][0] = v
            num_params = len(xr_baseparams)
            for j in range(num_params):
                data_list[row_offset+i][j+1] = "%g" % xr_baseparams[j]
                self.set_params_addr(xr_bp_base+j, (row_offset+i, j+1))

            # data_list[row_offset+i][xr_col_size+2] = "%g" % rgs[n+i]
            # self.set_params_addr(rg_base+n+i, (row_offset+i, xr_col_size+2))

        row_offset += 2
        data_list[row_offset][1] = "h"
        if uv_proportions is not None:
            data_list[row_offset][5] = "composite"
            data_list[row_offset][10] = "proportion"

        row_offset += 1
        if self.composite is None:
            proportions = uv_proportions
        else:
            proportions = self.composite.get_composite_proportions(uv_proportions)
        for i in range(n):
            if i == 0:
                v = "uv_params"
            else:
                v = ""
            data_list[row_offset+i][0] = v
            data_list[row_offset+i][1] = "%g" % uv_params[i]
            self.set_params_addr(uv_base+i, (row_offset+i, 1))

            if composite_ids:
                data_list[row_offset+i][xr_col_size+1] = str(composite_ids[i]+1)

            show_value = show_flags is None or show_flags[i]

            if uv_proportions is not None:
                if i == 0:
                    data_list[row_offset+i][xr_col_size+5] = "uv area"
                if i < len(uv_proportions):
                    if show_value:
                        i_ = i if composite_ids is None else composite_ids[i]
                        data_list[row_offset+i][xr_col_size+6] = "%g" % proportions[i_]

        row_offset += n + 1
        col_names = ["L", "x0", "k", "b", "s1", "s2", "diff_ratio"]
        if len(uv_baseparams) == 8:
            col_names.append("fouling")
        for j, name in enumerate(col_names):
            data_list[row_offset][1+j] = name

        row_offset += 1
        for i in range(1):
            if i == 0:
                v = "uv_baseline"
            else:
                v = ""
            data_list[row_offset+i][0] = v
            num_params = len(uv_baseparams)
            for j in range(num_params):
                data_list[row_offset+i][j+1] = "%g" % uv_baseparams[j]
                self.set_params_addr(uv_bp_base+j, (row_offset+i, j+1))

        row_offset += 2
        for j, name in enumerate(["slope", "intercept", "", "", "from", "to"]):
            data_list[row_offset][1+j] = name

        row_offset += 1
        data_list[row_offset][0] = "mapping"
        data_list[row_offset][4] = "mappable_range"
        for j in range(2):
            data_list[row_offset][1+j] = "%g" % mapping[j]
            self.set_params_addr(mp_base+j, (row_offset, 1+j))
            data_list[row_offset][5+j] = "%g" % mappable_range[j]
            self.set_params_addr(mr_base+j, (row_offset, 5+j))

        if seccol_params is not None:
            row_offset += 1
            for name, i in zip(["Npc", "rp (pore size)", "tI", "t0", "P", "m"], [0, 1, 2, 3, 4, 5]):
                value = seccol_params[i]
                row_offset += 1
                data_list[row_offset][0] = name
                data_list[row_offset][1] = "%g" % value
                self.set_params_addr(se_base+i, (row_offset, 1))

            if constrained_anyway and False:
                c9d_base = se_base + 4
                tmp_offset = row_offset - 3
                for i, (name, value) in enumerate([("Ti", Ti), ("Np", Np)]):
                    data_list[tmp_offset][3] = name
                    data_list[tmp_offset][4] = "%g" % value
                    self.set_params_addr(c9d_base + i, (tmp_offset, 4))
                    tmp_offset += 1

        self.num_valid_rows = row_offset + 2
        self.data_list = data_list
        column_width = 90
        width = column_width*num_columns + 60
        height = int(22*self.num_valid_rows) + 60
        self.sheet = Sheet(body_frame, width=width, height=height, data=data_list, show_selected_cells_border=False, column_width=column_width)

        for r, c in dark_cells:
            self.sheet.highlight_cells(r, c, bg="gray", fg="white")

        self.sheet.pack()



"""
    Protein.UvSensitivity.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import numpy as np
import matplotlib.pyplot as plt
from Bio.PDB import PDBParser, MMCIFParser, CaPPBuilder
from Bio.SeqUtils.ProtParam import ProteinAnalysis
from Bio.Data.IUPACData import protein_weights

pdb_parser = PDBParser()
cif_parser = MMCIFParser()
polypeptide_builder = CaPPBuilder()

def get_amino_acids_percent(pdb_file):
    parser = cif_parser if pdb_file[-4:] == '.cif' else pdb_parser
    structure = parser.get_structure("XXXX", pdb_file)

    result_dict = None
    for polypeptide in polypeptide_builder.build_peptides(structure):
        seq = polypeptide.get_sequence()
        analyzed_seq = ProteinAnalysis(str(seq))
        count_dict = analyzed_seq.count_amino_acids()
        if result_dict is None:
            result_dict = count_dict
        else:
            for k, v in count_dict.items():
                result_dict[k] += v

    # print(result_dict)
    total = np.sum(list(result_dict.values()))
    percent_dict = {}
    for k, v in result_dict.items():
        percent_dict[k] = v/total
    # print(percent_dict)

    return percent_dict

def get_ratio(proportion_dict):
    a280 = proportion_dict['W']
    a275 = proportion_dict['Y']
    a258 = proportion_dict['F']
    ratio = a280/a275 + a280/a258
    print("ratio=", ratio)
    return ratio

def plot_amino_acids_proportions(pdb_files, use_mass=False):
    fig, ax = plt.subplots()

    pdbids = []
    ratios = []
    for path in pdb_files:
        folder, file = os.path.split(path)
        pdbids.append(file.split("-")[0].upper())
        percent_dict = get_amino_acids_percent(path)
        if use_mass:
            method = "Mass"
            proportion_dict = {}
            total_mass = 0
            for k, v in percent_dict.items():
                mass = v*protein_weights[k]
                proportion_dict[k] = mass
                total_mass += mass
            for k, v in proportion_dict.items():
                proportion_dict[k] = v/total_mass
        else:
            method = "Count"
            proportion_dict = percent_dict
        ratio = get_ratio(proportion_dict)
        ratios.append(ratio)
        ax.bar(proportion_dict.keys(), proportion_dict.values(), alpha=0.3, label=file)

    ax.set_title("Amino Acid Proportions by %s: " % method + " vs ".join(pdbids))

    labels = ax.get_xticklabels()
    for i in [4, 18, 19]:
        labels[i].set_color("red")

    ax.legend()
    fig.tight_layout()

    print("FnR ratio = %.2g" % (ratios[1]/ratios[0]))

    plt.show()

def demo():
    import seaborn
    seaborn.set()

    DATA_FOLDER = r"D:\TODO\20231129\TDS\Data"
    paths = []
    # for file in ["2eph-ba1.pdb", "5f4x-ba1.pdb"]:
    for file in ["2eph-ba1.pdb", "3dfn-ba1.pdb"]:
        paths.append(os.path.join(DATA_FOLDER, file))

    plot_amino_acids_proportions(paths, use_mass=True)

if __name__ == '__main__':
    demo()



# coding: utf-8
"""
    EghBySampling.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from numpy.random import uniform
from scipy.interpolate import interp1d
from scipy.optimize import root
import matplotlib.pyplot as mplt
import molass_legacy.KekLib.DebugPlot as plt
from .EghMixture import egh_pdf, e1, e2, e3

class EghSampler:
    def __init__(self, tR=100, sigma=10, tau=10, start=0, stop=200):
        x = np.arange(start, stop, 0.5)
        y = egh_pdf(x, tR, sigma, tau)
        if False:
            plt.plot(x, y)
            plt.show()

        cdf_y = np.cumsum(y)
        cdf_y = cdf_y/cdf_y.max()
        if False:
            plt.plot(cdf_y, x)
            plt.show()
        if tau >= 0:
            self.inv_cdf = interp1d(cdf_y,x)
        else:
            pos_y = cdf_y > 0
            self.inv_cdf = interp1d(cdf_y[pos_y],x[pos_y])

    def __call__(self, num_samples):
        return  self.inv_cdf(np.random.rand(num_samples))

def moments_proof(tau_list):
    N = 1000000

    m1_list = []
    m2_list = []
    m3_list = []
    k3_list = []
    sigma = 10
    th_list = []
    tR = 100

    for tau in tau_list:
        print('tau=', tau)
        sampler = EghSampler(tR=tR, sigma=sigma, tau=tau)
        Z = sampler(N)
        th = np.arctan2(abs(tau), sigma)
        print('th=', th)
        th_list.append(th)
        M1 = np.mean(Z)
        m1 = tR + tau*e1(th)
        print('M1=', M1, m1)
        m1_list.append((M1, m1))
        M2 = np.sum((Z - M1)**2)/N
        m2 = (sigma**2 + sigma*abs(tau) + tau**2)*e2(th)
        print('M2=', M2, m2)
        m2_list.append((M2, m2))

        M3 = np.sum((Z - M1)**3)/N
        k3 = tau*(3*sigma**2 + 4*sigma*abs(tau) + 4*tau**2)
        m3 = k3*e3(th)
        print('M3=', M3, m3)
        k3_list.append(k3)
        m3_list.append((M3, m3))

    m1_array = np.array(m1_list)
    m2_array = np.array(m2_list)
    m3_array = np.array(m3_list)
    k3_array = np.array(k3_list)
    th_array = np.array(th_list)

    T = np.array([ th_array**k for k in range(7) ])
    Y = (m3_array[:,0]/k3_array).reshape((1, len(m3_array)))
    A = np.dot(Y, np.linalg.pinv(T))
    print('A=', A)

    x = tau_list

    plt.push()
    fig = plt.figure()
    ax = fig.gca()
    axt = ax.twinx()
    ax.set_title("Proof of EGH Moment Formulas with %d Sampling" % N, fontsize=16)
    ax.set_xlabel(r'$\tau$')
    ax.set_ylabel('Moments')
    ax.plot(x, m1_array[:,0], color='red', label=r'$M_1$ (raw)')
    ax.plot(x, m1_array[:,1], ':', color='red', label=r'$t_R + \tau\epsilon_1$')
    ax.plot(x, m2_array[:,0], color='green', label=r'$M_2$ (central)')
    ax.plot(x, m2_array[:,1], ':', color='green', label=r'$(\sigma^2 + \sigma|\tau| + \tau^2)\epsilon_2$')
    ax.plot(x, m3_array[:,0], color='blue', label=r'$M_3$ (central)')
    ax.plot(x, m3_array[:,1], ':', color='blue', label=r'$\tau(3\sigma^2 + 4\sigma|\tau| + 4\tau^2)\epsilon_3$')
    ax.legend()

    axt.plot(x, th_array, 'o', color='orange', label=r'$\Theta$')
    axt.legend(loc='center right')

    fig.tight_layout()
    plt.show()
    plt.pop()



"""
    Trimming.OptimalTrimming.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import logging
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from molass_legacy.UV.UvPreRecog import UvPreRecog

MARING_RATIO = 0.3      # relative to the peak region width
INCLUDING_FC_POS = -0.1

def get_start_stop_from_list(restrict_list, size):
    if restrict_list is None:
        e_start = 0
        e_stop = size
    else:
        etrim = restrict_list[0]
        if etrim is None:
            e_start = 0
            e_stop = size
        else:
            e_start = 0 if etrim.start is None else etrim.start
            e_stop = size if etrim.stop is None else etrim.stop
    return e_start, e_stop

def including_fc_pos(f, t):
    return f*(1+INCLUDING_FC_POS) - t*INCLUDING_FC_POS

class OptimalTrimming:
    def __init__(self, sd, pre_recog, debug=False):
        self.logger = logging.getLogger(__name__)
        if debug:
            self.sd = sd
        else:
            self.sd = None
        peak_region_width = pre_recog.cs.a_curve.get_peak_region_width()

        upr = UvPreRecog(sd, pre_recog, debug=debug)
        trim_slice = upr.get_trim_slice()
        if hasattr(sd, 'conc_array'):
            # for SerialData
            uv_size = sd.conc_array.shape[1]
        else:
            # for SecSaxs.DataSet
            uv_size = sd.uv_array.shape[1]
        ut0_ = uv_size if trim_slice.stop is None else trim_slice.stop

        uv_restrict_list = get_setting('uv_restrict_list')
        uv_e_start, uv_e_stop = get_start_stop_from_list(uv_restrict_list, uv_size)
        xr_restrict_list = get_setting('xr_restrict_list')
        xr_size = sd.intensity_array.shape[0]
        xr_e_start, xr_e_stop = get_start_stop_from_list(xr_restrict_list, xr_size)

        similarity = pre_recog.cs
        mp_e_start, mp_e_stop = [ similarity.mapped_int_value(j) for j in [xr_e_start, xr_e_stop] ]
        ut0_ = min(ut0_, mp_e_stop)

        # make sure to include flowchange point if one exists near the peak region as in 20220311
        fc = pre_recog.flowchange.get_real_flow_changes()
        fc_left = fc[0]
        if fc_left is None:
            fc_left = 0

        inc_pos = including_fc_pos(uv_e_start, uv_e_stop)
        if fc_left is None or fc_left < inc_pos:
            start = max(uv_e_start, mp_e_start)
        else:
            margin = int(peak_region_width*MARING_RATIO)
            f0 = max(0, fc_left - margin)
            start = min(f0, uv_e_start)

        peak_slice = pre_recog.cs.a_curve.peak_slice
        if (peak_slice.stop - peak_slice.start)/uv_size < 0.8:  # almost the same meaning with peak_slice.stop - peak_slice.start < uv_size
            uf0 = min(peak_slice.start, start)
            ut0 = max(peak_slice.stop, ut0_)
        else:
            uf0 = start
            ut0 = ut0_

        xf0, xt0 = [similarity.inverse_int_value(j) for j in [uf0, ut0]]
        xr_size = sd.intensity_array.shape[0]
        xf1 = max(0, xf0)
        xt1 = min(xr_size, xt0)

        self.uv_ends = [uf0, ut0]
        self.xr_ends = [xf1, xt1]
        self.base_curve_info = upr.get_base_curve_info()

        if debug:
            from molass_legacy.Elution.CurveUtils import simple_plot
            print([uf0, ut0], [xf0, xt0], [xf1, xt1])
            curves = sd.get_elution_curves()
            uv_base_curve, init_params = self.base_curve_info

            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("debug plot at OptimalTrimming")
                simple_plot(ax1, curves[0])
                x = curves[0].x
                y = curves[0].y
                ax1.plot(x, uv_base_curve(x, init_params, y), color="red")
                ymin, ymax = ax1.get_ylim()
                ax1.set_ylim(ymin, ymax)

                for j in fc:
                    if j is not None:
                        ax1.plot([j, j], [0, 0], "o", color="cyan")

                for j in [uv_e_start, uv_e_stop]:
                    ax1.plot([j, j], [ymin, ymax], ":", color="gray")

                for j in [uf0, ut0]:
                    ax1.plot([j, j], [ymin, ymax], color="yellow")

                for j in [peak_slice.start, peak_slice.stop]:
                    ax1.plot([j, j], [ymin, ymax], ":", color="green")

                simple_plot(ax2, curves[1], color="orange")
                ymin, ymax = ax2.get_ylim()
                ax2.set_ylim(ymin, ymax)
                for j in [xf1, xt1]:
                    ax2.plot([j, j], [ymin, ymax], color="yellow")
                fig.tight_layout()
                plt.show()

    def get_base_curve_info(self, debug=False):
        uv_base_curve, init_params = self.base_curve_info
        if debug and self.sd is not None:
            from molass_legacy.Elution.CurveUtils import simple_plot
            uv_curve = self.sd.get_uv_curve()
            x = uv_curve.x
            y = uv_curve.y
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("get_base_curve_info debug")

                simple_plot(ax1, uv_curve)
                by1 = uv_base_curve(x, init_params, y)
                ax1.plot(x, by1, color="red")

                simple_plot(ax2, uv_curve)
                # by2 = shifted_curve(x, shifted_params)
                # ax2.plot(x, by2, color="red")

                fig.tight_layout()
                plt.show()

        return uv_base_curve, init_params

    def set_info(self, debug=False):
        uv_restrict_list = get_setting("uv_restrict_list")
        xr_restrict_list = get_setting("xr_restrict_list")

        if debug:
            print("uv_restrict_list=", uv_restrict_list)
            print("xr_restrict_list=", xr_restrict_list)

        if uv_restrict_list is not None:
            uv_info = uv_restrict_list[0]
            if uv_info is not None:
                uv_info.start = self.uv_ends[0]
                uv_info.stop = self.uv_ends[1]

        if xr_restrict_list is not None:
            xr_info = xr_restrict_list[0]
            if xr_info is not None:
                xr_info.start = self.xr_ends[0]
                xr_info.stop = self.xr_ends[1]

        if debug:
            print("uv_restrict_list=", uv_restrict_list)
            print("xr_restrict_list=", xr_restrict_list)
        set_setting("uv_restrict_list", uv_restrict_list)
        self.logger.info("uv_restrict_list has been set to %s in OptimalTrimming", str(uv_restrict_list))
        set_setting("xr_restrict_list", xr_restrict_list)

def spike(in_folder, logger, debug=False):
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition

    set_setting("in_folder", in_folder)
    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)

    pre_recog = PreliminaryRecognition(sd)
    mt = OptimalTrimming(sd, pre_recog, debug=debug)
    mt.set_info(debug=True)
